/* -*- linux-c -*-
 * FILE NAME board.h
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *           board layer API
 *
 *
 * Author:   Matthias Fuchs
 *           matthias.fuchs@esd-electronics.com
 *
 * history:
 *
 *  27.05.02 - first version                                       mf
 *  11.05.04 - changed OSIF_EXTERN into extern and OSIF_CALLTYPE   ab
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file board.h
    \brief Card interface API

    This file contains the API for accessing the CAN driver's card layer.
    The functions are called by the nucleus modules.
*/

#ifndef __BOARD_H__
#define __BOARD_H__

#include <esdcan.h>
#include <osif.h>
#include <cm.h>
#include <canio.h>

extern INT32 cif_open( CAN_OCB *ocb, CAN_NODE *node, INT32 flags );
extern INT32 cif_close( CAN_OCB *ocb );
extern INT32 cif_node_attach( CAN_NODE *node );
extern INT32 cif_node_detach( CAN_NODE *node );
extern INT32 cif_id_filter( CAN_NODE *node, UINT32 cmd, UINT32 id, UINT32 *count );
extern OSIF_CALLTYPE INT32 cif_tx( CAN_NODE *node );
extern INT32 cif_tx_obj( CAN_NODE *node, UINT32 cmd, VOID *arg );
extern INT32 cif_tx_abort( CAN_NODE *node, CM *cm, INT32 status );
extern INT32 cif_baudrate_set( CAN_NODE *node, UINT32 baud );

extern INT32 cif_ioctl( CAN_OCB *ocb, VOID *waitContext, UINT32 cmd, 
                        VOID *buf_in, UINT32 len_in, VOID *buf_out, UINT32 *len_out );

extern INT32 cif_statistics( CAN_NODE *node, CAN_STAT *stat );
extern INT32 cif_timestamp( CAN_NODE *node, CAN_TS *timestamp, CAN_TS_FREQ *frequency );

/* 1) implementation for soft timestamps is in board/<board>/board.c
   2) implementation for hard timestamps is in board/<board>/boardrc.c
      (this is very hardware specific !)
*/
extern INT32 cif_softts_get( VOID *dummy, CAN_TS *ts );
extern INT32 cif_softts_freq_get( VOID *dummy, CAN_TS_FREQ *ts_freq );

/* implemented for intelligent cards, only */
extern OSIF_CALLTYPE INT32 cif_debug( CAN_NODE* node, CAN_DEBUG_ARG *can_debug );


OSIF_CALLTYPE INT32 can_board_attach( CAN_CARD *crd );
OSIF_CALLTYPE INT32 can_board_attach_final( CAN_CARD *crd );
OSIF_CALLTYPE INT32 can_board_detach( CAN_CARD *crd );

#endif /* __BOARD_H__ */



