/* -*- linux-c -*-
 * FILE NAME boardrc.h
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *           board resources for pci200 boards
 *
 *
 * Author:   Matthias Fuchs
 *           matthias.fuchs@esd-electronics.com
 *
 * history:
 *  $Log: boardrc.h,v $
 *  Revision 1.14  2006/10/12 09:47:08  andreas
 *  Added support for PCIe2000 with new subsystem ID (0x200)
 *
 *  Revision 1.13  2005/07/28 13:22:25  andreas
 *  PCI defines changed slightly, when adding cardFlavours array and
 *   implemented hotplugable PCI driver.
 *  BOARD_STRING removed, ESDCAN_DRIVER_NAME added.
 *  Board type is defined here, now (BOARD_PCI).
 *
 *  Revision 1.12  2005/06/22 07:35:25  oliver
 *  New define IDX_BOARD for CAN-PCI/200
 *
 *  Revision 1.11  2005/04/20 11:26:21  andreas
 *  Fixed PCI-structure (use of standard C-types exchanged with OSIF-types) for 64-Bit systems
 *  Fixed pragma pack for gcc 2.95.x (open with #pragma pack(1), but close with #pragma pack())
 *
 *  Revision 1.10  2004/11/22 16:04:34  matthias
 *  added smart disconnect mode to feature flags
 *
 *
 *  11.05.04 - added _ID_ suffix to PCI-IDs (for kernel 2.6)
 *             changed OSIF_EXTERN to extern (s. OSIF_CALLTYPE)  ab
 *  15.07.02 - first version                                     mf
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file boardrc.h
    \brief Board specific board APi

    This file contains board specific constants for a particular board implementation.
    This file may be empty for some boards.
*/

#ifndef __BOARDRC_H__
#define __BOARDRC_H__

#include <osif.h>
#include <cm.h>

#ifndef OSIF_KERNEL
#error "This file may be used in the kernel-context, only! Not for application-use!!!"
#endif

#define ESDCAN_DRIVER_NAME "CAN_PCI2xx"

/* default driver features - more feature flags will be set after hardware probing */
#define CARD_FEATURES          (FEATURE_CAN_20B | FEATURE_LOM | FEATURE_TS | FEATURE_SMART_DISCONNECT)

#define BOARD_PCI

#define NODES_PER_CARD                2
#define MAX_CARDS                     4

#define SJA1000_CLOCK                 16000000

#define PCI_VENDOR_PLX                0x10B5
#define PCI_SUB_VENDOR_ESD            0x12FE

#define PCI_DEVICE_PLX9050            0x9050
#define PCI_DEVICE_PLX9030            0x9030
#define PCI_DEVICE_PLX9056            0x9056

#define PCI_SUB_SYSTEM_PCI200         0x0004
#define PCI_SUB_SYSTEM_D32            0x0003
#define PCI_SUB_SYSTEM_CPCI200        0x010B
#define PCI_SUB_SYSTEM_PCI266         0x0009
#define PCI_SUB_SYSTEM_D32_66         0x000A
#define PCI_SUB_SYSTEM_PCIE2000       0x0200

#define PCI_PLX_ICS_LI1S_9050_9030    0x00000004
#define PCI_PLX_ICS_LI1EN_9050_9030   0x00000001
#define PCI_PLX_ICS_PIEN_9050_9030    0x00000040
#define PCI_PLX_ICS_OFFSET_9050_9030  0x4C

#define PCI_PLX_ICS_LI1S_9056         0x00008000
#define PCI_PLX_ICS_LI1EN_9056        0x00000800
#define PCI_PLX_ICS_PIEN_9056         0x00000100
#define PCI_PLX_ICS_OFFSET_9056       0x68

#define PCI_TYPE0_ADDRESSES           6

#define ESDCAN_IDS_PCI200 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9050, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_PCI200, \
        0, 0, 0 }
#define ESDCAN_IDS_D32 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9050, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_D32, \
        0, 0, 0 }
#define ESDCAN_IDS_CPCI200 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9030, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_CPCI200, \
        0, 0, 0 }
#define ESDCAN_IDS_PCI266 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9056, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_PCI266, \
        0, 0, 0 }
#define ESDCAN_IDS_D32_66 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9056, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_D32_66, \
        0, 0, 0 }
#define ESDCAN_IDS_PCIE2000 { \
        PCI_VENDOR_PLX, PCI_DEVICE_PLX9056, \
        PCI_SUB_VENDOR_ESD, PCI_SUB_SYSTEM_PCIE2000, \
        0, 0, 0 }


#define ESDCAN_IDS_PCI ESDCAN_IDS_PCI200, ESDCAN_IDS_CPCI200, ESDCAN_IDS_PCI266, \
                       ESDCAN_IDS_D32, ESDCAN_IDS_D32_66, ESDCAN_IDS_PCIE2000

extern CARD_IDENT cardFlavours[];

#pragma pack(1)
/* address space 0 of (C)PCI200 */
struct  PCI200_CAN_CFG
{
        UINT8  unused[PCI_PLX_ICS_OFFSET_9050_9030];
        UINT32 ics;           /* interrupt-control status */
        UINT32 misc_ctrl;     /* user-io/pci-target-response/eeprom/init-control */
};
/* address space 0 of PCI266/D32 */
struct  PCI266_CAN_CFG
{
        UINT8  unused[PCI_PLX_ICS_OFFSET_9056];
        UINT32 ics;           /* interrupt-control status */
        UINT32 misc_ctrl;     /* user-io/pci-target-response/eeprom/init-control */
};
#pragma pack()

#define CIF_CARD \
	OSIF_PCIDEV  pciDev;

#endif
