/* canio.h
**
**            Copyright (c) 2001 by electronic system design gmbh
**
**  This software is copyrighted by and is the sole property of
**  esd gmbh.  All rights, title, ownership, or other interests
**  in the software remain the property of esd gmbh. This
**  software may only be used in accordance with the corresponding
**  license agreement.  Any unauthorized use, duplication, transmission,
**  distribution, or disclosure of this software is expressly forbidden.
**
**  This Copyright notice may not be removed or modified without prior
**  written consent of esd gmbh.
**
**  esd gmbh, reserves the right to modify this software without notice.
**
**  electronic system design gmbh          Tel. +49-511-37298-0
**  Vahrenwalder Str 205                   Fax. +49-511-37298-68
**  30165 Hannover                         http://www.esd-electronics.com
**  Germany                                slaes@esd-electronics.com
**
**
** user-header for  io-control for esd-can-driver linux
**
*/
#ifndef __CANIO_H__
#define __CANIO_H__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <linux/ioctl.h>

/* Following define is used to en-/disable code on non-64-Bit platforms */
#define PLATFORM_64BIT  defined(__ppc64__) || defined(__x86_64__)

#ifdef __KERNEL__
# define CANIO_UINT8   u8
# define CANIO_INT8    s8
# define CANIO_UINT16  u16
# define CANIO_INT16   s16
# define CANIO_UINT32  u32
# define CANIO_INT32   s32
# define CANIO_UINT64  u64
# define CANIO_INT64   s64
#else
# include <stdint.h>
# define CANIO_UINT8   uint8_t
# define CANIO_INT8    int8_t
# define CANIO_UINT16  uint16_t
# define CANIO_INT16   int16_t
# define CANIO_UINT32  uint32_t
# define CANIO_INT32   int32_t
# define CANIO_UINT64  uint64_t
# define CANIO_INT64   int64_t
#endif

#define CANIO_20B_BASE          0x20000000

/* CAN-id-values (11-Bit/20a)   */
#define CANIO_IDS               0x800      /* range of 20a-IDs */
#define CANIO_ID_BASE           0x00000000
#define CANIO_ID_LAST           CANIO_ID_BASE+CANIO_IDS-1

/* CAN-id-values (29-Bit/20b)   */
#define CANIO_IDS_20B           0x20000000 /* range of 20b-IDs */
#define CANIO_ID_20B_BASE       0x20000000
#define CANIO_ID_20B_LAST       CANIO_ID_20B_BASE+CANIO_IDS_20B-1

#define CANIO_ID_INVALID        0xFFFFFFFF

/* CAN-event-id-values          */
#define CANIO_EVS               0x100      /* range of events */
#define CANIO_EV_BASE           0x40000000
#define CANIO_EV_CAN_ERROR      CANIO_EV_BASE
#define CANIO_EV_BAUD_CHANGE    CANIO_EV_BASE + 0x01
#define CANIO_EV_CAN_DEFERRED_TX CANIO_EV_BASE+ 0x30
#define CANIO_EV_ECHO(NR)       CANIO_EV_BASE + 0x40 + (NR & 0x1f)
#define CANIO_EV_ECHO_FIRST     CANIO_EV_BASE + 0x40
#define CANIO_EV_ECHO_LAST      CANIO_EV_BASE + 0x5F
#define CANIO_EV_USER           CANIO_EV_BASE + 0x80
#define CANIO_EV_DNET_EVPOLL    CANIO_EV_BASE + 0x7c
#define CANIO_EV_DNET_EVEXPL    CANIO_EV_BASE + 0x7d
#define CANIO_EV_DNET_EV        CANIO_EV_BASE + 0x7e
#define CANIO_EV_DNET_CALL      CANIO_EV_BASE + 0x7f
#define CANIO_EV_WATCHDOG       CANIO_EV_USER + 0x05
#define CANIO_EV_WATCHDOG_RESET CANIO_EV_USER + 0x45
#define CANIO_EV_LAST           CANIO_EV_BASE + 0xFF

#define CANIO_DN_BASE           0x41000000
#define CANIO_DN_LAST           0x41ffffff

#define CANIO_RTR               0x00000010
#define CANIO_RTR_MASK          0x0010
#define CANIO_NO_DATA_MASK      0x0020  /* only senseful in obj-mode */
#define CANIO_INTERACTION       0x0020  /* only senseful in non obj-mode */
#define CANIO_ERROR_MASK        0x0040

#define CANIO_AUTO_ANSWER_VALID 0x0020  /* AB added */

#define CANIO_MAX_TX_QUEUESIZE  0x4000
#define CANIO_MAX_RX_QUEUESIZE  0x4000

#define CANIO_MAX_TX_TIMEOUT    0xFFFF
#define CANIO_MAX_RX_TIMEOUT    0xFFFF

#define CANIO_BAUD_NOT_SET      0x7FFFFFFF   /* unset baudrate(nuc_baudrate_set/get */
#define CANIO_USER_BAUD         0x80000000
#define CANIO_LISTEN_ONLY_MODE  0x40000000
#define CANIO_USER_BAUD_NUM     0x20000000
#define CANIO_AUTOBAUD          0x7FFFFFFE

/* Defines to decode CAN controller type, as delivered by canStatus(), etc. */
#define CANIO_CANCTL_SJA1000    0x00
#define CANIO_CANCTL_I82527     0x01
#define CANIO_CANCTL_FUJI       0x02
#define CANIO_CANCTL_LPC2292    0x03
#define CANIO_CANCTL_MSCAN      0x04


/*********************************************************************
 * IO-controls                                                       *
 *********************************************************************/
/*
 * Define own private IOCTL for "candev"-driver
 */
#define IOCTL_CAN_SET_ID          1
#define IOCTL_CAN_SET_BAUD        2
#define IOCTL_CAN_SET_MODE        3
#define IOCTL_CAN_SET_TIMEOUT     4
#define IOCTL_CAN_SEND            5
#define IOCTL_CAN_TAKE            7
#define IOCTL_CAN_UPDATE          8
#define IOCTL_CAN_SET_QUEUESIZE   9  /* Also used to distinguish old and new driver */
#define IOCTL_CAN_TX_BUFFER       10
#define IOCTL_CAN_RX_BUFFER       11
#define IOCTL_CAN_DEBUG           12
#define IOCTL_CAN_GET_BAUD        17
#define IOCTL_CAN_SEND_DEVNET     18
#define IOCTL_CAN_FLUSH_RX_FIFO   19
#define IOCTL_CAN_GET_RX_MESSAGES 20
#define IOCTL_CAN_GET_RX_TIMEOUT  21
#define IOCTL_CAN_GET_TX_TIMEOUT  22
#define IOCTL_CAN_ABORT           23
#define IOCTL_CAN_RAW             250


/*
 * Structures used as args for "candev"-ioctls
 */
/*
 * IOCTL_CAN_SET_BAUD:
 *  arg = CAN_BAUDSET*
 *  Values for baud-argument:
 *  baud = int 0   = 1000.0 kBit/s
 *             1   =  666.6 kBit/s
 *             2   =  500.0 kBit/s
 *             3   =  333.3 kBit/s
 *             4   =  250.0 kBit/s
 *             5   =  166.0 kBit/s
 *             6   =  125.0 kBit/s
 *             7   =  100.0 kBit/s
 *             8   =   66.6 kBit/s
 *             9   =   50.0 kBit/s
 *             10  =   33.3 kBit/s
 *             11  =   20.0 kBit/s
 *             12  =   12.5 kBit/s
 *             13  =   10.0 kBit/s
 *             14  =  800.0 kBit/s
 *             15  = 1600.0 kBit/s
 *             16  =   83.3 kBit/s
 */
typedef struct          /* structure for baud setting */
{
     int32_t net;           /* net number                 */
     int32_t baud;          /* baudrate                   */
} CAN_BAUDSET, *PCAN_BAUDSET;


typedef union {
        uint64_t          tick;
#if ( defined(__BYTE_ORDER) && (__BYTE_ORDER == __BIG_ENDIAN) ) || ( !defined(__BYTE_ORDER) && defined(__BIG_ENDIAN))
        struct {
                uint32_t  HighPart;
                uint32_t  LowPart;
        } h;
#endif /* __BIG_ENDIAN */
#if ( defined(__BYTE_ORDER) && (__BYTE_ORDER == __LITTLE_ENDIAN) ) || ( !defined(__BYTE_ORDER) && defined(__LITTLE_ENDIAN))
        struct {
                uint32_t  LowPart;
                uint32_t  HighPart;
        } h;
#endif /* __LITTLE_ENDIAN */
} CAN_TIMESTAMP, CAN_TIMESTAMP_FREQ;


/*
 * arg = CAN_MSG*
 */
typedef struct {              /* structure for action-routines */
        int32_t  id;          /* id of can-frame or event-id   */
        uint8_t  len;         /* count of data-bytes (0...8)   */
        uint8_t  msg_lost;    /* message-lost-counter          */
        uint8_t  reserved[2]; /* reserved for future           */
        union {               /* data-buffer                   */
                int8_t   c[8];
                int16_t  s[4];
                int32_t  l[2];
                int64_t  ll[1];
        } buf;
} CAN_MSG, *PCAN_MSG;

typedef struct {
        int32_t       id;             /* can-id                                   */
        uint8_t       len;            /* length of message: 0-8                   */
        uint8_t       msg_lost;       /* count of lost rx-messages                */
        uint8_t       reserved[2];    /* reserved                                 */
        uint8_t       data[8];        /* 8 data-bytes                             */
        CAN_TIMESTAMP timestamp;      /* time stamp of this message               */
} CAN_MSG_T, *PCAN_MSG_T;


typedef struct _CAN_SCHED
{
	int32_t  id;
	int32_t  flags;
	uint64_t time_start;
	uint64_t time_interval;
	uint8_t  reserved[8];
} CAN_SCHED;


/*
 * IOCTL_CAN_SET_QUEUESIZE:
 * arg = CAN_QUEUE*
 */
typedef struct {
        int32_t  tx_size; /* queuesize for tx-direction */
        int32_t  rx_size; /* queuesize for rx-direction */
} CAN_QUEUE, *PCAN_QUEUE;

#pragma pack(1)
typedef struct {        /* structure for action-routines   */
        uint32_t  id;   /* 0x40/0x01/macid/len of dnet req */
        uint16_t  x1;   /* para2                           */
        uint16_t  x2;   /* para2                           */
        union {         /* data-buffer                     */
                int8_t   c[8];
                int16_t  s[4];
                int32_t  l[2];
                int64_t  ll[1];
        } buf;
} DNET_MSG, *PDNET_MSG;
#pragma pack()



typedef struct {
    int32_t  id1st;
    int32_t  idlst;
} CAN_IDSET;


/*
 * Define own private IOCTL for "esdcan"-driver
 * AB: To esd developers:
 *     Don't forget to add the "_32-defines" (for 64-platforms)
 *     and additions to esdcan_unregister_ioctl32() and
 *     esdcan_register_ioctl32() in esdcan.c, when adding new IOCTLs.
 */
#define	IOCTL_BASE  'C'

#define IOCTL_ESDCAN_CREATE                    _IOW  (IOCTL_BASE,  0, IOCTL_ARG*)
#define IOCTL_ESDCAN_ID_RANGE_ADD              _IOW  (IOCTL_BASE,  1, IOCTL_ARG*)
#define IOCTL_ESDCAN_ID_RANGE_DEL              _IOW  (IOCTL_BASE,  2, IOCTL_ARG*)
#define IOCTL_ESDCAN_SET_BAUD                  _IOW  (IOCTL_BASE,  3, IOCTL_ARG*)
#define IOCTL_ESDCAN_SET_MODE                  _IOW  (IOCTL_BASE,  4, IOCTL_ARG*) /* AB: DEPRECATED/REMOVED, not needed in esdcan (candev legacy)*/
#define IOCTL_ESDCAN_SET_TIMEOUT               _IOW  (IOCTL_BASE,  5, IOCTL_ARG*)
#define IOCTL_ESDCAN_SEND                      _IOW  (IOCTL_BASE,  6, IOCTL_ARG*)
#define IOCTL_ESDCAN_TAKE                      _IOR  (IOCTL_BASE,  7, IOCTL_ARG*)
#define IOCTL_ESDCAN_DEBUG                     _IOWR (IOCTL_BASE,  9, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_BAUD                  _IOR  (IOCTL_BASE, 10, IOCTL_ARG*)
#define IOCTL_ESDCAN_DESTROY                   _IO   (IOCTL_BASE, 11)
#define IOCTL_ESDCAN_DESTROY_DEPRECATED        _IOW  (IOCTL_BASE, 11, unsigned long) /* Compatibility to libs <= 3.0.7 */
#define IOCTL_ESDCAN_TX_ABORT                  _IOW  (IOCTL_BASE, 12, IOCTL_ARG*)
#define IOCTL_ESDCAN_RX_ABORT                  _IOW  (IOCTL_BASE, 13, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_CREATE             _IOW  (IOCTL_BASE, 16, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_UPDATE             _IOW  (IOCTL_BASE, 17, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_OFF     _IOW  (IOCTL_BASE, 18, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_DESTROY            _IOW  (IOCTL_BASE, 19, IOCTL_ARG*)
#define IOCTL_ESDCAN_RX_OBJ                    _IOW  (IOCTL_BASE, 20, IOCTL_ARG*)
#define IOCTL_ESDCAN_READ                      _IOWR (IOCTL_BASE, 21, IOCTL_ARG*)
#define IOCTL_ESDCAN_WRITE                     _IOWR (IOCTL_BASE, 22, IOCTL_ARG*)
#define IOCTL_ESDCAN_UPDATE                    _IOW  (IOCTL_BASE, 23, IOCTL_ARG*)
#define IOCTL_ESDCAN_FLUSH_RX_FIFO             _IO   (IOCTL_BASE, 24)
#define IOCTL_ESDCAN_FLUSH_RX_FIFO_DEPRECATED  _IOW  (IOCTL_BASE, 24, unsigned long) /* Compatibility to libs <= 3.0.7 */
#define IOCTL_ESDCAN_GET_RX_MESSAGES           _IOR  (IOCTL_BASE, 25, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_RX_TIMEOUT            _IOR  (IOCTL_BASE, 26, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_TX_TIMEOUT            _IOR  (IOCTL_BASE, 27, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_TIMESTAMP_FREQ        _IOR  (IOCTL_BASE, 28, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_TIMESTAMP             _IOR  (IOCTL_BASE, 29, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_TICKS_FREQ            _IOR  (IOCTL_BASE, 30, IOCTL_ARG*) /* AB: DEPRECATED/REMOVED, use IOCTL_ESDCAN_GET_TIMESTAMP_FREQ */
#define IOCTL_ESDCAN_GET_TICKS                 _IOR  (IOCTL_BASE, 31, IOCTL_ARG*) /* AB: DEPRECATED/REMOVED, use IOCTL_ESDCAN_GET_TIMESTAMP */
#define IOCTL_ESDCAN_SEND_T                    _IOW  (IOCTL_BASE, 32, IOCTL_ARG*)
#define IOCTL_ESDCAN_WRITE_T                   _IOW  (IOCTL_BASE, 33, IOCTL_ARG*)
#define IOCTL_ESDCAN_TAKE_T                    _IOWR (IOCTL_BASE, 34, IOCTL_ARG*)
#define IOCTL_ESDCAN_READ_T                    _IOWR (IOCTL_BASE, 35, IOCTL_ARG*)
#define IOCTL_ESDCAN_PURGE_TX_FIFO             _IO   (IOCTL_BASE, 36)
#define IOCTL_ESDCAN_PURGE_TX_FIFO_DEPRECATED  _IOW  (IOCTL_BASE, 36, unsigned long) /* Compatibility to libs <= 3.0.7 */
#define IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_ON      _IOW  (IOCTL_BASE, 37, IOCTL_ARG*)
#define IOCTL_ESDCAN_SET_RX_TIMEOUT            _IOW  (IOCTL_BASE, 38, IOCTL_ARG*)
#define IOCTL_ESDCAN_SET_TX_TIMEOUT            _IOW  (IOCTL_BASE, 39, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_FEATURES              _IOR  (IOCTL_BASE, 40, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE           _IOW  (IOCTL_BASE, 41, IOCTL_ARG*)
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE_START     _IO   (IOCTL_BASE, 42)
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE_STOP      _IO   (IOCTL_BASE, 43)
#define IOCTL_ESDCAN_SET_20B_HND_FILTER        _IOW  (IOCTL_BASE, 44, IOCTL_ARG*)
#define IOCTL_ESDCAN_GET_SERIAL                _IOR  (IOCTL_BASE, 45, IOCTL_ARG*)
#define IOCTL_ESDCAN_SET_ALT_RTR_ID            _IO   (IOCTL_BASE, 46)

#if PLATFORM_64BIT
#define GEN_IOCTL32(arg) (((arg) & (~IOCSIZE_MASK)) | (0x4 << IOCSIZE_SHIFT)) /* patch pointer size difference in ioctl commands */

#define REGISTER_IOCTL_32(arg, ret) \
        if ( ret |= register_ioctl32_conversion(arg##_32, esdcan_ioctl_32) ) { \
                CAN_PRINT(("esd CAN driver: Problem with registering IOCTL32: 0x%x!\n", arg##_32)); \
        }
#define REGISTER_IOCTL_COMPAT(arg, ret) \
        if ( ret |= register_ioctl32_conversion(arg, NULL) ) { \
                CAN_PRINT(("esd CAN driver: Problem with registering IOCTL32: 0x%x!\n", arg)); \
        }
#define UNREGISTER_IOCTL_32(arg) \
        if ( unregister_ioctl32_conversion(arg##_32) ) { \
                CAN_PRINT(("esd CAN driver: Problem with unregistering IOCTL32: 0x%x!\n", arg##_32)); \
        }
#define UNREGISTER_IOCTL_COMPAT(arg) \
        if ( unregister_ioctl32_conversion(arg) ) { \
                CAN_PRINT(("esd CAN driver: Problem with unregistering IOCTL32: 0x%x!\n", arg)); \
        }

#define IOCTL_ESDCAN_CREATE_32                     GEN_IOCTL32(IOCTL_ESDCAN_CREATE)
#define IOCTL_ESDCAN_ID_RANGE_ADD_32               GEN_IOCTL32(IOCTL_ESDCAN_ID_RANGE_ADD)
#define IOCTL_ESDCAN_ID_RANGE_DEL_32               GEN_IOCTL32(IOCTL_ESDCAN_ID_RANGE_DEL)
#define IOCTL_ESDCAN_SET_BAUD_32                   GEN_IOCTL32(IOCTL_ESDCAN_SET_BAUD)
#define IOCTL_ESDCAN_SET_MODE_32                   GEN_IOCTL32(IOCTL_ESDCAN_SET_MODE)
#define IOCTL_ESDCAN_SET_TIMEOUT_32                GEN_IOCTL32(IOCTL_ESDCAN_SET_TIMEOUT)
#define IOCTL_ESDCAN_SEND_32                       GEN_IOCTL32(IOCTL_ESDCAN_SEND)
#define IOCTL_ESDCAN_TAKE_32                       GEN_IOCTL32(IOCTL_ESDCAN_TAKE)
#define IOCTL_ESDCAN_DEBUG_32                      GEN_IOCTL32(IOCTL_ESDCAN_DEBUG)
#define IOCTL_ESDCAN_GET_BAUD_32                   GEN_IOCTL32(IOCTL_ESDCAN_GET_BAUD)
#define IOCTL_ESDCAN_DESTROY_32                    /* same as IOCTL_ESDCAN_DESTROY */
#define IOCTL_ESDCAN_DESTROY_DEPRECATED_32         GEN_IOCTL32(IOCTL_ESDCAN_DESTROY_DEPRECATED)
#define IOCTL_ESDCAN_TX_ABORT_32                   GEN_IOCTL32(IOCTL_ESDCAN_TX_ABORT)
#define IOCTL_ESDCAN_RX_ABORT_32                   GEN_IOCTL32(IOCTL_ESDCAN_RX_ABORT)
#define IOCTL_ESDCAN_TX_OBJ_CREATE_32              GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_CREATE)
#define IOCTL_ESDCAN_TX_OBJ_UPDATE_32              GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_UPDATE)
#define IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_OFF_32      GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_OFF)
#define IOCTL_ESDCAN_TX_OBJ_DESTROY_32             GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_DESTROY)
#define IOCTL_ESDCAN_RX_OBJ_32                     GEN_IOCTL32(IOCTL_ESDCAN_RX_OBJ)
#define IOCTL_ESDCAN_READ_32                       GEN_IOCTL32(IOCTL_ESDCAN_READ)
#define IOCTL_ESDCAN_WRITE_32                      GEN_IOCTL32(IOCTL_ESDCAN_WRITE)
#define IOCTL_ESDCAN_UPDATE_32                     GEN_IOCTL32(IOCTL_ESDCAN_UPDATE)
#define IOCTL_ESDCAN_FLUSH_RX_FIFO_32              /* same as IOCTL_ESDCAN_PURGE_TX_FIFO */
#define IOCTL_ESDCAN_FLUSH_RX_FIFO_DEPRECATED_32   GEN_IOCTL32(IOCTL_ESDCAN_FLUSH_RX_FIFO_DEPRECATED)
#define IOCTL_ESDCAN_GET_RX_MESSAGES_32            GEN_IOCTL32(IOCTL_ESDCAN_GET_RX_MESSAGES)
#define IOCTL_ESDCAN_GET_RX_TIMEOUT_32             GEN_IOCTL32(IOCTL_ESDCAN_GET_RX_TIMEOUT)
#define IOCTL_ESDCAN_GET_TX_TIMEOUT_32             GEN_IOCTL32(IOCTL_ESDCAN_GET_TX_TIMEOUT)
#define IOCTL_ESDCAN_GET_TIMESTAMP_FREQ_32         GEN_IOCTL32(IOCTL_ESDCAN_GET_TIMESTAMP_FREQ)
#define IOCTL_ESDCAN_GELT_TIMESTAMP_32              GEN_IOCTL32(IOCTL_ESDCAN_GET_TIMESTAMP)
#define IOCTL_ESDCAN_GET_TICKS_FREQ_32             GEN_IOCTL32(IOCTL_ESDCAN_GET_TICKS_FREQ)
#define IOCTL_ESDCAN_GET_TICKS_32                  GEN_IOCTL32(IOCTL_ESDCAN_GET_TICKS)
#define IOCTL_ESDCAN_SEND_T_32                     GEN_IOCTL32(IOCTL_ESDCAN_SEND_T)
#define IOCTL_ESDCAN_WRITE_T_32                    GEN_IOCTL32(IOCTL_ESDCAN_WRITE_T)
#define IOCTL_ESDCAN_TAKE_T_32                     GEN_IOCTL32(IOCTL_ESDCAN_TAKE_T)
#define IOCTL_ESDCAN_READ_T_32                     GEN_IOCTL32(IOCTL_ESDCAN_READ_T)
#define IOCTL_ESDCAN_PURGE_TX_FIFO_32              /* same as IOCTL_ESDCAN_PURGE_TX_FIFO */
#define IOCTL_ESDCAN_PURGE_TX_FIFO_DEPRECATED_32   GEN_IOCTL32(IOCTL_ESDCAN_PURGE_TX_FIFO_DEPRECATED)
#define IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_ON_32       GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_AUTOANSWER_ON)
#define IOCTL_ESDCAN_SET_RX_TIMEOUT_32             GEN_IOCTL32(IOCTL_ESDCAN_SET_RX_TIMEOUT)
#define IOCTL_ESDCAN_SET_TX_TIMEOUT_32             GEN_IOCTL32(IOCTL_ESDCAN_SET_TX_TIMEOUT)
#define IOCTL_ESDCAN_GET_FEATURES_32               GEN_IOCTL32(IOCTL_ESDCAN_GET_FEATURES)
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE_32            GEN_IOCTL32(IOCTL_ESDCAN_TX_OBJ_SCHEDULE)
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE_START_32      /* same as IOCTL_ESDCAN_TX_OBJ_SCHEDULE_START */
#define IOCTL_ESDCAN_TX_OBJ_SCHEDULE_STOP_32       /* same as IOCTL_ESDCAN_TX_OBJ_SCHEDULE_STOP */
#define IOCTL_ESDCAN_SET_20B_HND_FILTER_32         GEN_IOCTL32(IOCTL_ESDCAN_SET_20B_HND_FILTER)
#define IOCTL_ESDCAN_GET_SERIAL_32                 GEN_IOCTL32(IOCTL_ESDCAN_GET_SERIAL)
#define IOCTL_ESDCAN_SET_ALT_RTR_ID_32             /* same as IOCTL_ESDCAN_SET_ALT_RTR_ID */

#endif

/*
 * Structures used as args for "esdcan"-ioctls
 */
typedef struct {
	void     *buffer;
	uint32_t  size;
} IOCTL_ARG;

#if PLATFORM_64BIT
typedef struct {
	uint32_t  buffer;
	uint32_t  size;
} IOCTL_ARG_32;
#endif

/* argument for CAN_INIT ioctl */
typedef struct {
#ifdef OSIF_OS_RTAI
	uint32_t  net_no;
	int32_t   hnd;
#endif
	uint32_t  mode;
	uint32_t  queue_size_tx;
	uint32_t  queue_size_rx;
} CAN_INIT_ARG;

/* argument for CAN_ID_ADD/DEL ioctl */
typedef struct {
	uint32_t  id_start;
	uint32_t  id_stop;
} CAN_ID_RANGE_ARG;


/*
 * Structures used as args for "esdcan"- and "candev"-ioctls
 */
/*
 * IOCTL_CAN_SET_TIMEOUT, IOCTL_ESDCAN_SET_TIMEOUT:
 * arg = CAN_TIMEOUT*
 */
typedef struct {         /* structure for timeout-setting */
        uint32_t txtout; /* tx-timeout in msec            */
        uint32_t rxtout; /* rx-timeout in msec            */
} CAN_TIMEOUT, *PCAN_TIMEOUT, CAN_TIMEOUT_ARG;

/*
 * IOCTL_CAN_DEBUG, IOCTL_ESDCAN_DEBUG:
 * arg = CAN_DEBUG*
 */
#pragma pack(1)
typedef struct {                               /* structure for action-routines */
        uint32_t  password;                    /* password                      */
        uint32_t  cmd;                         /* command to execute            */
        union {
                struct {
                        uint8_t   access;      /* byte/word/long access */
                        uint8_t   bus;         /* bus-nr                */
                        uint8_t   reserved[2];
                        uint32_t  address;
                        union {
                                uint8_t   c;
                                uint16_t  s;
                                uint32_t  l;
                        } data;
                } dm_sm;
                struct {
                        uint16_t  drv_ver;
                        uint16_t  firm_ver;
                        uint16_t  hard_ver;
                        uint16_t  board_stat;
                        int8_t    firm_info[14];
                        uint16_t  features;
                        int8_t    reserved[6];
                } ver;
                struct {
                        uint32_t  sector;
                        uint16_t  offset;
                        uint16_t  len;
                        uint8_t   data[14];
                } flash;
                struct {
                        int8_t mode;
                } trans_mode;
        } p;
} CAN_DEBUG, *PCAN_DEBUG, CAN_DEBUG_ARG;
#pragma pack()

typedef struct
{
        uint32_t  baud;         /* value configured by user via canSetBaudrate() */
        uint32_t  valid;        /* validity of all _following_ infos (-1 = invalid, CANIO_SUCCESS, CANIO_NOT_IMPLEMENTED) */
        uint32_t  rate;         /* CAN bitrate in Bit/s */
        uint32_t  clock;        /* frequency of CAN controller */
        uint32_t  sp;           /* in percent */
        uint32_t  error;        /* in percent */
        uint8_t   ctrl_type;    /* CANIO_CANCTL_XXX defines */
        uint8_t   reserved1[3]; /* for future use and alignment */
        uint32_t  reserved2;    /* for future use */
        uint32_t  reserved3;    /* for future use */
        uint32_t  baud_req;     /* requested bitrate, while request pending, else -1 (RESERVED IN NTCAN) */
        union {
                struct {
                        uint8_t   btr0;
                        uint8_t   btr1;
                        uint8_t   tseg1;
                        uint8_t   tseg2;
                        uint8_t   brp;
                        uint8_t   sjw;
                        uint8_t   sam;
                } sja1000; /* also for: i82527 and MSCAN */
                struct {
                        uint16_t  btr;
                        uint8_t   tseg1;
                        uint8_t   tseg2;
                        uint8_t   psc;
                        uint8_t   sjw;
                        uint8_t   sam;
                } fuji;
                struct {
                        uint32_t  btr;
                        uint8_t   tseg1;
                        uint8_t   tseg2;
                        uint16_t  brp;
                        uint8_t   sjw;
                        uint8_t   sam;
                } lpc2292;
        } ctrl; /* infos stored in bit timing register(s) of controller */
} CAN_BITRATE;

/*
 *  Mode-flags (prefix CANIO_MODE_) are used by user/application
 *  with IOCTL_ESDCAN_SET_MODE to set the driver into a certain mode.
 *  !!! They have to be defined according to ntcan.h !!!
 */
#define CANIO_MODE_AUTO_ANSWER       0x00000002 /* mode-mask for auto-answer         */
#define CANIO_MODE_NO_RTR            0x00000010 /* mode-mask for ignoring rtr        */
#define CANIO_MODE_NO_DATA           0x00000020 /* mode-mask for ignoring data       */
#define CANIO_MODE_SET_RTR_RCV       0x00000040 /* used to mark active RTR-reception (TODO: ???) */
#define CANIO_MODE_RTR_RCV           0x00000060 /* mode-mask for object mode         (TODO: ???) */
#define CANIO_MODE_NO_INTERACTION    0x00000100 /* Ignore locally send interaction messages */
#define CANIO_MODE_MARK_INTERACTION  0x00000200 /* Mark interaction messages in len field */
#define CANIO_MODE_LOCAL_ECHO        0x00000400 /* Echo sent frames to same handle   */
/* reserved NTCAN_MODE_SPS_CNTRL     0x00008000    Enable firmware SPS             */
#define CANIO_MODE_LOM               0x00010000 /* mode-mask for listening only mode */
#define CANIO_MODE_DEFERRED_TX       0x00020000 /* mode-mask for "scheduled tx"      */
#define CANIO_MODE_OBJECT            0x10000000 /* mode-mask for object mode         */
#define CANIO_MODE_OVERLAPPED        0x20000000 /* reserved for windows, only        */
#define CANIO_MODE_EVENTS            0x40000000 /* mode-mask for enabled events (TODO: ???) */


/*
 *  Feature flags for "candev"-driver
 */
#define FEATURE_FULL_CAN         (1<<0)  /*! Marks Full-CAN-controller, used by nucleus for flat-mode,... */
#define FEATURE_CAN_20B          (1<<1)  /*! CAN 2.OB support                                             */
#define FEATURE_DEVICE_NET       (1<<2)
#define FEATURE_RX_OBJECT_MODE   (1<<4)  /*! DIFFERS FROM MODE FLAG: flat (receive object) mode           */
/* AB: the following feature-flags aren't implemented in candev-driver */
#define FEATURE_TS               (1<<5)  /*! feature flag for hardware timestamping                       */
#define FEATURE_LOM              (1<<6)  /*! feature/mode flag for listening-only-mode                    */
#define FEATURE_SMART_DISCONNECT (1<<7)  /*! feature/mode flag for stay-on-bus-after-last-close           */
#define FEATURE_LOCAL_ECHO       (1<<8)  /*! feature/mode interaction with local echo                     */

#define FEATURE_MASK_DOCUMENTED  (FEATURE_FULL_CAN |         \
                                  FEATURE_CAN_20B |          \
                                  FEATURE_DEVICE_NET |       \
                                  FEATURE_RX_OBJECT_MODE |   \
                                  FEATURE_TS |               \
                                  FEATURE_SMART_DISCONNECT | \
                                  FEATURE_LOM )

/*
 *  Following defines are used in a nodes status cell for
 *    - i20: wrong firmware / hardware
 *    - bus status (ok, warn, err. passive, bus off)
 */
#define CANIO_STATUS_WRONG_FIRMWARE  0x00000001  /*! Wrong firmware version (disables CAN-functionality) */
#define CANIO_STATUS_WRONG_HARDWARE  0x00000002  /*! Wrong hardware version (disables CAN-functionality) */
#define CANIO_STATUS_OK              0x00000000  /*! can Status = OK                                     */
#define CANIO_STATUS_WARN            0x40000000  /*! can status = controller warn                        */
#define CANIO_STATUS_ERRPASSIVE      0x80000000  /*! can status = error passive                          */
#define CANIO_STATUS_BUSOFF          0xC0000000  /*! can status = controller off bus                     */

#define CANIO_STATUS_MASK_BUSSTATE   0xC0000000  /*! use to retrieve bus states                       */

/*
 * Debug-command-values
 */
#define CAN_DEBUG_DM            0   /* dump memory */
#define CAN_DEBUG_SM            1   /* set memory  */
#define CAN_DEBUG_IP            2   /* input port  */
#define CAN_DEBUG_OP            3   /* output port */
#define CAN_DEBUG_VER           4   /* get version-information */
/*#define CAN_DEBUG_MAP         5   */    /* map memory-space to user-space */
/*#define CAN_DEBUG_REGISTER   10   */    /* display register */
/*#define CAN_DEBUG_RXFIFO     11   */    /* display interface rx fifo */
#define CAN_DEBUG_FLASH_ENTER 128   /* enter flash-mode */
#define CAN_DEBUG_FLASH_ERASE 129   /* erase sector */
#define CAN_DEBUG_FLASH_PROG  130   /* flash some bytes */
#define CAN_DEBUG_FLASH_EXIT  131   /* exit flash-mode  */
#define CAN_DEBUG_FLASH_PROG1 132   /* flash some bytes */
#define CAN_DEBUG_SET_TRANS   140   /* Set firmware mode to 2.0A or 2.0B */

/* debug-access-values */
#define CAN_ACCESS_BYTE 0
#define CAN_ACCESS_WORD 1
#define CAN_ACCESS_LONG 2


/*
 * can-error-codes
 */
#define CANIO_ERRNO_BASE             0x00000100  /* base for esd-codes */

#define CANIO_SUCCESS                0

#define CANIO_RX_TIMEOUT             (CANIO_ERRNO_BASE+1)
#define CANIO_TX_TIMEOUT             (CANIO_ERRNO_BASE+2)
/* #define CANIO_UNUSED              (CANIO_ERRNO_BASE+3) */
#define CANIO_TX_ERROR               (CANIO_ERRNO_BASE+4)
#define CANIO_CONTR_OFF_BUS          (CANIO_ERRNO_BASE+5)
#define CANIO_CONTR_BUSY             (CANIO_ERRNO_BASE+6)
#define CANIO_CONTR_WARN             (CANIO_ERRNO_BASE+7)
#define CANIO_OLDDATA                (CANIO_ERRNO_BASE+8)
#define CANIO_NO_ID_ENABLED          (CANIO_ERRNO_BASE+9)
#define CANIO_ID_ALREADY_ENABLED     (CANIO_ERRNO_BASE+10)
#define CANIO_ID_NOT_ENABLED         (CANIO_ERRNO_BASE+11)
/* #define CANIO_UNUSED              (CANIO_ERRNO_BASE+12) */
#define CANIO_INVALID_FIRMWARE       (CANIO_ERRNO_BASE+13)
#define CANIO_MESSAGE_LOST           (CANIO_ERRNO_BASE+14)
#define CANIO_INVALID_HARDWARE       (CANIO_ERRNO_BASE+15)
#define CANIO_PENDING_WRITE          (CANIO_ERRNO_BASE+16)
#define CANIO_PENDING_READ           (CANIO_ERRNO_BASE+17)
#define CANIO_INVALID_DRIVER         (CANIO_ERRNO_BASE+18)

#define CANIO_INVALID_PARAMETER      EINVAL
#define CANIO_INVALID_HANDLE         EBADFD
/* #define CANIO_IO_INCOMPLETE       not reasonable under Linux */
/* #define CANIO_IO_PENDING          not reasonable under Linux */
#define CANIO_NET_NOT_FOUND          ENODEV
#define CANIO_INSUFFICIENT_RESOURCES ENOMEM

#define CANIO_OPERATION_ABORTED      EINTR
#define CANIO_WRONG_DEVICE_STATE     (CANIO_ERRNO_BASE+19)
#define CANIO_HANDLE_FORCED_CLOSE    (CANIO_ERRNO_BASE+20)
#define CANIO_NOT_IMPLEMENTED        ENOSYS
#define CANIO_NOT_SUPPORTED          (CANIO_ERRNO_BASE+21)
#define CANIO_CONTR_ERR_PASSIVE      (CANIO_ERRNO_BASE+22)

/* Errors used in esdcan, but not formerly present in canio.h
 * Used in situations, where errno-systems return the respecting errno */
#define CANIO_EIO                    EIO
#define CANIO_EFAULT                 EFAULT
#define CANIO_EBUSY                  EBUSY
#define CANIO_EAGAIN                 EAGAIN
#define CANIO_ENOENT                 ENOENT

/* This error should never ever reach user level!!! */
#define CANIO_ERROR                  1  /* todo: Generic Error...not EPERM */

#ifdef OSIF_OS_RTAI
INT32 esdcan_ctrl(INT32 hnd, UINT32 cmd, UINT32 arg);
INT32 esdcan_write(INT32 hnd, VOID *buf, INT32 count);
INT32 esdcan_read(INT32 hnd, VOID *buf, INT32 count);
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* __CANIO_H__ */
