/* -*- linux-c -*-
 * FILE NAME cm.h
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *
 *
 *
 * history:
 *
 *  $Log: cm.h,v $
 *  Revision 1.33  2006/07/11 15:12:41  manuel
 *  Added ts and ts_freq to CAN_STAT
 *
 *  Revision 1.32  2006/06/27 09:53:56  andreas
 *  Added  ESDCAN_CTL_BAUDRATE_AUTO and ESDCAN_CTL_BAUDRATE_BTR
 *  Added baud to CAN_STAT (quick fix in order to remove nuc_baudrate_get())
 *
 *  Revision 1.31  2005/09/29 07:22:47  michael
 *  internal ioctl-codes added
 *
 *  Revision 1.30  2005/09/14 13:25:36  manuel
 *  Added CM_LEN2DATALEN macro
 *
 *  Revision 1.29  2005/07/27 15:53:01  andreas
 *  Added CARD_IDENT structure (see some boardrc.c/h for usage info)
 *
 *  27.05.02 - first version                         mf
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file cm.h
 *  \brief Contains common CAN-message-structure (CM) and defines such as
 *  feature-flags, mode-flags. These defines are visible to the user,
 *  nevertheless: Dear user, please, use the defines in ntcan.h!
 *
 */

#ifndef __CM_H__
#define __CM_H__

#include <osif.h>

#ifndef OSIF_KERNEL
#error "This file may be used in the kernel-context, only! Not for application-use!!!"
#endif

#define CM_RTR           0x00000010
#define CM_NO_DATA       0x00000020

/* calculate the data length from a given frame length value */
#define CM_LEN2DATALEN(l) (((l)&0x10)?0:(((l)>8)?8:(l)))

/* command-flags for nuc_id_filter                                           */
#define FILTER_ON         0x00000000      /* enable flag                     */
#define FILTER_OFF        0x80000000      /* disable flag                    */

#define FILTER_DATA       0x00000001      /* enabled for queued data input   */
#define FILTER_RTR        0x00000002      /* enabled for queued rtr input    */
#define FILTER_OBJ        0x00000004      /* enabled for object data input   */
#define FILTER_OBJ_RTR    0x00000008      /* enabled for object rtr input    */
#define FILTER_ALL (FILTER_DATA | FILTER_RTR | FILTER_OBJ | FILTER_OBJ_RTR)


/* Following is defined for the board files but deprecated */
#define FILTER_DATA_ENABLE    FILTER_DATA
#define FILTER_RTR_ENABLE     FILTER_RTR
#define FILTER_OBJ_ENABLE     FILTER_OBJ
#define FILTER_OBJ_RTR_ENABLE FILTER_OBJ_RTR_ENABLE
#define FILTER_DISABLE        0x00000000

#define ESDCAN_CTL_TIMESTAMP_GET      1
#define ESDCAN_CTL_TIMEST_FREQ_GET    2
#define ESDCAN_CTL_BAUDRATE_GET       3
#define ESDCAN_CTL_BAUDRATE_SET       4
#define ESDCAN_CTL_ID_FILTER          5
#define ESDCAN_CTL_BAUDRATE_AUTO      6
#define ESDCAN_CTL_BAUDRATE_BTR       7

typedef union
  {
    UINT64 tick;
    struct
    {
#ifdef OSIF_LITTLE_ENDIAN
      UINT32 LowPart;
      UINT32 HighPart;
#endif
#ifdef OSIF_BIG_ENDIAN
      UINT32 HighPart;
      UINT32 LowPart;
#endif
    } h;
} CAN_TS, CAN_TS_FREQ;

typedef	union {
	UINT64 ul64[1];
	UINT32 ul32[2];
	VOID*  ptr;
} HOST_HND;


typedef struct _CMSCHED CMSCHED;
struct _CMSCHED
{
	UINT32 id;
	INT32  flags;
	CAN_TS timeStart;
	CAN_TS timeInterval;
	UINT8  reserved[8];
};


typedef struct _CM CM;
struct _CM {
	UINT32        id;             /* can-id                                   */
	UINT8         len;            /* length of message: 0-8                   */
	UINT8         msg_lost;       /* count of lost rx-messages                */
	UINT8         reserved[2];    /* reserved                                 */
	UINT8         data[8];        /* 8 data-bytes                             */
        CAN_TS        timestamp;      /* 64 bit timestamp                         */
	HOST_HND      host_hnd;
};

typedef struct _VERSION {
	UINT32        level;
	UINT32        revision;
	UINT32        build;
} VERSION;

/* test-only */
typedef struct _CAN_STAT {
	UINT32      msg_count_a[8+1]; /* CAN 2.0A messages */
	UINT32      msg_count_b[8+1]; /* CAN 2.0B messages */
        UINT32      baud;             /* baudrate in bit/s */
	UINT32      byte_count;
	CAN_TS      ts;
	CAN_TS_FREQ ts_freq;
} CAN_STAT;

typedef struct _CARD_IRQ {
	OSIF_IRQ_HANDLER( *handler, context);
	VOID *context;
} CARD_IRQ;

/* Use this to terminate cardFlavours arrays */
#define CARD_IDENT_PCI_TERMINATE {{ NULL, NULL, NULL, {0, 0, 0, 0, 0, 0, 0} }}
#define CARD_IDENT_RAW_TERMINATE {{ NULL, NULL, NULL }}
#define CARD_IDENT_USB_TERMINATE {{ NULL, NULL, NULL, {0, 0} }}

typedef union _CARD_IDENT {
        struct {
                CARD_IRQ   *irqs;
                UINT32     *spaces;
                const INT8 *name;
                struct {
                        UINT32   vendor;
                        UINT32   device;
                        UINT32   subVendor;
                        UINT32   subDevice;
                        UINT32   class;     /* For future use, init with 0    */
                        UINT32   classMask; /* For future use, init with 0    */
                        UINTPTR  misc;      /* For future use, init with NULL */
                } ids;
        } pci;
        struct {
                CARD_IRQ   *irqs;
                UINT32     *spaces;
                const INT8 *name;
        } raw;
        struct {
                CARD_IRQ   *irqs;
                UINT32     *spaces;
                const INT8 *name;
                struct {
                        UINT32   vendor;
                        UINT32   device;
                } ids;
        } usb;
        struct { /* This part needs to be identical for all kinds of boards */
                CARD_IRQ   *irqs;
                UINT32     *spaces;
                const INT8 *name;
        } all;
} CARD_IDENT;

#define MAKE_VERSION(l,r,b) (((l) << 12) | ((r) << 8) | ((b) << 0))

#endif  /* #ifndef __CM_H__ */
