/* -*- linux-c -*-
 * FILE NAME nucleus.h
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *           ...
 *
 *
 * Author:   Matthias Fuchs
 *           matthias.fuchs@esd-electronics.com
 *
 * history:
 *
 *  07.03.06 - added nuc_deferred_tx_messages
 *           - added nuc_tx_messages                                   mk
 *  11.05.04 - changed OSIF_EXTERN into extern and OSIF_CALLTYPE       ab
 *  13.11.03 - added nuc_get_ocb()
 *             renamed nuc_tx_abort_force() into nuc_tx_abort_board()  ab
 *  24.05.02 - first version                                           mf
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2006 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file nucleus.h
 *  \brief Contains defines and prototypes of the CAN-nucleus.
 *
 *  \par General rules:
 *
 */

#ifndef __NUCLEUS_H__
#define __NUCLEUS_H__

#include <osif.h>
#include <esdcan.h>
#include <cm.h>

#ifndef OSIF_KERNEL
#error "This file may be used in the kernel-context, only! Not for application-use!!!"
#endif

#define NUC_NODE_RX_QUEUE_SIZE   1024         /* rx queue inside node */
#define NUC_OCB_PER_ID      15

#define NUC_TX_OBJ_CREATE         0
#define NUC_TX_OBJ_AUTOANSWER_ON  1
#define	NUC_TX_OBJ_AUTOANSWER_OFF 2
#define NUC_TX_OBJ_UPDATE         3
#define NUC_TX_OBJ_DESTROY        4
#define	NUC_TX_OBJ_SCHEDULE       5
#define	NUC_TX_OBJ_SCHEDULE_START 6
#define	NUC_TX_OBJ_SCHEDULE_STOP  7

#define NUC_RX_OBJ_CREATE    0
#define NUC_RX_OBJ_TAKE      1
#define NUC_RX_OBJ_DESTROY   2

extern UINT32 baud_table[];

typedef struct _NUC_CALLBACKS NUC_CALLBACKS;
struct _NUC_CALLBACKS {
	OSIF_CALLTYPE INT32 (*close_done)( CAN_OCB *ocb );
	OSIF_CALLTYPE INT32 (*tx_done)( CAN_OCB *ocb, INT32 result, UINT32 cm_count );
	OSIF_CALLTYPE INT32 (*tx_get)( CAN_OCB *ocb, CM *cm );
	OSIF_CALLTYPE INT32 (*tx_obj_get)( CAN_OCB *ocb, CM *cm );
	OSIF_CALLTYPE INT32 (*tx_obj_sched_get)( CAN_OCB *ocb, CMSCHED *sched );
	OSIF_CALLTYPE INT32 (*rx_done)( CAN_OCB *ocb, INT32 result, UINT32 cm_count );
	OSIF_CALLTYPE INT32 (*rx_put)( CAN_OCB *ocb, CM *cm );
	OSIF_CALLTYPE INT32 (*rx_obj_put)( CAN_OCB *ocb, CM *cm, UINT32 *next_id );
	OSIF_CALLTYPE INT32 (*ioctl_done)( CAN_OCB *ocb, VOID *waitContext, UINT32 cmd, INT32 result, UINT32 len_out );
/*	INT32 (*err_put)( CAN_NODE *can_node, INT32 err ); */
};

/*
 * nucleus interface for OS module
 */
extern OSIF_CALLTYPE INT32 nuc_open( CAN_OCB      *ocb,
                                     CAN_NODE     *node,
                                     NUC_CALLBACKS *cm,
                                     UINT32       flags,
                                     UINT32       queue_size_tx,
                                     UINT32       queue_size_rx );

extern OSIF_CALLTYPE INT32 nuc_close( CAN_OCB *ocb );
extern OSIF_CALLTYPE INT32 nuc_tx_abort( CAN_OCB *ocb, INT32 status, UINT64 host_hnd );
extern OSIF_CALLTYPE INT32 nuc_rx_abort( CAN_OCB *ocb, INT32 status );
extern OSIF_CALLTYPE INT32 nuc_rx_purge( CAN_OCB *ocb );
extern OSIF_CALLTYPE INT32 nuc_rx_messages( CAN_OCB *ocb, UINT32 *count );
extern OSIF_CALLTYPE INT32 nuc_tx_messages( CAN_OCB *ocb, UINT32 *count );
extern OSIF_CALLTYPE INT32 nuc_deferred_tx_messages( CAN_OCB *ocb, UINT32 *count );


extern OSIF_CALLTYPE INT32 nuc_ioctl( CAN_OCB *can_ocb, VOID *waitContext, UINT32 cmd,
                                      VOID *buf_in, UINT32 len_in, VOID *buf_out, UINT32 *len_out );

extern OSIF_CALLTYPE INT32 nuc_rx( CAN_OCB *ocb,  UINT32 timeout, UINT32 *count );
extern OSIF_CALLTYPE INT32 nuc_rx_obj( CAN_OCB *ocb, UINT32 *count, UINT32 id );
extern OSIF_CALLTYPE INT32 nuc_tx( CAN_OCB *ocb,  UINT32 timeout, UINT32 *count );
extern OSIF_CALLTYPE INT32 nuc_tx_obj( CAN_OCB *ocb, UINT32 cmd, UINT32 *count );

extern OSIF_CALLTYPE INT32 nuc_node_attach( CAN_NODE *node );
extern OSIF_CALLTYPE INT32 nuc_node_detach( CAN_NODE *node );
extern OSIF_CALLTYPE INT32 nuc_baudrate_set(CAN_NODE *node, UINT32 baud);
extern OSIF_CALLTYPE INT32 nuc_baudrate_get(CAN_NODE *node, UINT32 *baud);
extern OSIF_CALLTYPE INT32 nuc_id_filter(CAN_OCB *ocb, UINT32 cmd, UINT32 id, UINT32 *count );
extern OSIF_CALLTYPE INT32 nuc_id_range_check(UINT32 id_1st, UINT32 id_lst);

extern OSIF_CALLTYPE INT32 nuc_statistics( CAN_NODE *node, CAN_STAT *stat );
extern OSIF_CALLTYPE INT32 nuc_extra_info( CAN_NODE *node, UINT8 *buf, INT32 count, INT32 *len);
extern OSIF_CALLTYPE INT32 nuc_timestamp( CAN_NODE *node, CAN_TS *timestamp, CAN_TS_FREQ *frequency );

extern OSIF_CALLTYPE CAN_OCB *nuc_get_ocb( CAN_NODE *can_node, CAN_OCB *ocb );

/*
 * nucleus interface for CIF module
 */

extern OSIF_CALLTYPE VOID nuc_tx_get( CAN_NODE *node, CM **cm );
extern OSIF_CALLTYPE INT32 nuc_tx_done( CAN_NODE *node, INT32 status, CM *cm );
extern OSIF_CALLTYPE INT32 nuc_rx_put( CAN_NODE *node,  CM **cm );
/* extern OSIF_CALLTYPE INT32 nuc_err_put( CAN_NODE *node, INT32 error ); */
extern OSIF_CALLTYPE VOID  nuc_tx_abort_board( INT32 status, CM *cm );
extern OSIF_CALLTYPE INT32 nuc_ioctl_done( CAN_OCB *can_ocb, VOID *waitContext, UINT32  cmd, INT32 status, INT32 len_out );

#endif










