/* -*- linux-c -*-
 * FILE NAME osif.c
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *	     OSIF implementation for Linux
 *
 *
 *
 * history:
 *
 *  $Log: osif.c,v $
 *  Revision 1.80  2006/10/12 13:31:55  manuel
 *  Fixed compiler warning
 *
 *  Revision 1.79  2006/10/11 10:12:17  andreas
 *  Fixed compilation problem with SuSE > 10.0
 *
 *  Revision 1.78  2006/08/17 13:26:59  michael
 *  Rebirth of OSIF_ errorcodes
 *
 *  Revision 1.77  2006/08/03 07:46:19  andreas
 *  Removed old debug code, which was #if 0 for quite a while, now
 *
 *  Revision 1.76  2006/06/29 15:18:40  andreas
 *  CANIO_EINTR got replaced by OSIF_OPERATION_ABORTED
 *
 *  Revision 1.75  2006/06/29 14:45:06  andreas
 *  Exchanged error codes from osif.h (OSIF_xxx) with error codes from canio.h (CANIO_xxx)
 *
 *  Revision 1.74  2006/06/27 13:13:44  andreas
 *  Exchanged OSIF_errors with CANIO_errors
 *
 *  Revision 1.73  2006/06/27 09:52:03  andreas
 *  TODO comment added (regarding timestamps on baudrate change events)
 *
 *  Revision 1.72  2006/04/28 07:54:27  andreas
 *  osif_ticks() must not be used with mecp52 (or generally mpc5200 based boards)
 *
 *  Revision 1.71  2006/02/14 09:18:37  matthias
 *  -board info structure has change to 2.6 kernels
 *  -check if _MSC_VER is defined be fore using it
 *
 *  Revision 1.70  2005/12/06 13:24:10  andreas
 *  No functional changes, only cleanup, while looking through the
 *   code in search for an error
 *
 *  Revision 1.69  2005/10/06 07:43:23  matthias
 *  added osif_make_ntcan_serial
 *
 *  Revision 1.68  2005/08/30 12:25:09  andreas
 *  Fixed osif_memcmp() !!!
 *  This one was really nice (great copy'n'paste work), comparison was done,
 *  result was piped to NIL.
 *  Instead the result was always ok :(
 *
 *  Revision 1.67  2005/07/27 15:50:14  andreas
 *  Corrected calltype of func-parameter of osif_timer_create
 *
 *  Revision 1.66  2005/04/20 13:04:16  andreas
 *  Changes for 64-Bit Linux (exchanged C-standard types with OSIF-types, use of UINTPTR)
 *  Added osif_ser_out()
 *
 *  Revision 1.65  2004/11/15 17:35:54  matthias
 *  fixed frequency of PPC750 performance counter frequency
 *
 *  Revision 1.64  2004/11/15 17:25:14  matthias
 *  added temporary osif_tick_frequency() for cpci750
 *
 *  Revision 1.63  2004/11/08 11:13:46  michael
 *  Missing OSIF_CALLTYPE added
 *
 *  Revision 1.62  2004/10/29 14:56:05  matthias
 *  -added osif_div64_64
 *  -fixed osif_timer_set rounding
 *
 *  Revision 1.61  2004/10/11 09:23:44  andreas
 *  Added io-functions (for kernel 2.6)
 *  Added some debug zones
 *  Sorted history
 *
 *  13.05.04 - corrected SMP-bug in OSIF_SLEEP for kernel < 2.6
 *             new OSIF_SLEEP branch for kernel > 2.6.5
 *             timer functions needed OSIF_CALLTYPE, too
 *             changed dpc_thread kernel2.6-style (daemonize()...)  ab
 *  11.05.04 - Changed OSIF_EXTERN into OSIF_CALLTYPE
 *             Added rcsid                                          ab
 *  07.05.04 - Removed gcc3-warning (LVALUE-casts)
 *             Including linux/version.h
 *             Changed prototype of osif_irq_attach a bit to cope
 *               with the needs of kernel 2.6                       ab
 *  06.05.04 - Include of modversions changed for kernel 2.6.x
 *             Added OSIF_EXTERN to nearly all functions (problem
 *             with release-archive for kernel 2.6.x)               ab
 *  01.12.03 - OSIF doesn't use OSIF-macros for mutex/sema anymore  mt
 *  26.11.03 - Corrected OSIF_VSNPRINTF for kernel < 2.4.0
 *             Removed osif_pci_malloc-warning for non-pci405-cards
 *             Exchanged init_MUTEX with sema_init()                ab
 *  13.11.03 - Corrected debug-info in mutexes                      ab
 *  13.11.03 - OSIF_MUTEX do not busy wait, Old OSIF_MUTEX now
 *             OSIF_BUSY_MUTEX                                      mt
 *  12.11.03 - OSIF_TIMER as DPC                                    mt
 *  05.11.03 - Saving of IRQ-flags was wrong! Removed again!        ab
 *  29.10.03 - changed IRQ-spinlocks (IRQ-flags get saved, now)     ab
 *  27.10.03 - New output function osif_dprint (with verbose mask)  ab
 *  18.09.03 - Only one thread for all DPC's                        mt
 *  02.01.03 - added SA_INTERRUPT flag for atomic interrupt handler mf
 *  24.05.02 - first version                                        mf
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file osif.c
    \brief OSIF implementation

    This file contains the implementation of all osif functions.
*/

#include <linux/config.h>
#include <linux/version.h>

#ifdef DISTR_SUSE
# undef CONFIG_MODVERSIONS
#endif
#ifdef CONFIG_MODVERSIONS
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#  include <config/modversions.h>
# else
#  include <linux/modversions.h>
# endif
# ifndef MODVERSIONS
#  define MODVERSIONS
# endif
#endif

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <asm/semaphore.h>
#include <linux/timer.h>

#ifdef LTT
# include <linux/trace.h>
#endif

#include <linux/errno.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
# include <linux/interrupt.h>
#endif

#include <canio.h>
#include <osif.h>


#if ( (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,8)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)) )
/* needed for vsnprintf */
#include <asm/div64.h>
#include <linux/ctype.h>
#endif

#ifdef ESDDBG
# define OSIF_DEBUG                       /*!< enables debug output for osif functions */
#endif

#ifdef OSIF_DEBUG
# define OSIF_DBG(fmt)	        OSIF_DPRINT(fmt) /*!< Macro for debug prints */
#else
# define OSIF_DBG(fmt)
#endif

struct __OSIF_SEMA {
	struct semaphore sema;
};

struct __OSIF_SPINLOCK {
	spinlock_t spinlock;
	UINT32     magic;
};

struct __OSIF_TIMER {
	struct timer_list   timer;
	OSIF_DPC            dpc;
};

struct __OSIF_PCI_PADDR {
	UINT32              addr32;
	dma_addr_t          sys_dma_addr;
};

extern UINT32 verbose;  /* has to be declared in OS-level-sources (esdcan.c), should be set by module-parameter */

static INT32 osif_dpc_root_create( VOID);
static INT32 osif_dpc_root_destroy( VOID);

static struct {
	OSIF_DPC            first;
	spinlock_t          spinlock;
	INT32               thread_running;
	struct semaphore    sema_notify;
	struct semaphore    sema_trigger;
	INT32               destroy;
} dpc_root;

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
DECLARE_COMPLETION(dpcThreadCompletion);
#endif


OSIF_CALLTYPE INT32 osif_attach()
{
        INT32 result;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	result = osif_dpc_root_create();

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return result;
}

OSIF_CALLTYPE INT32 osif_detach()
{
        INT32 result;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	result = osif_dpc_root_destroy();

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return result;
}

/*! Allocate \a size bytes of memory. A pointer to the new allocated memory is
  returned through \a **p.
*/
#define MEM_IS_KMEM 0
#define MEM_IS_VMEM 1
OSIF_CALLTYPE INT32 osif_malloc(UINT32 size, VOID **p)
{
	VOID  *ptr;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter, size=%08x\n", OSIF_FUNCTION,size));

	ptr = kmalloc(size + sizeof(UINT32), GFP_KERNEL);
	if (!ptr) {
		OSIF_DBG((OSIF_ZONE_OSIF, "%s: kmalloc for %d bytes failed, using vmalloc\n", OSIF_FUNCTION, size));
		ptr = vmalloc(size + sizeof(UINT32));
		if (!ptr) {
			*p = NULL;
			OSIF_DBG((OSIF_ZONE_OSIF, "%s: OSIF_MALLOC failed !!!!\n", OSIF_FUNCTION));
			return OSIF_INSUFFICIENT_RESOURCES;
		} else {
			UINT32 *p_long = (UINT32*)ptr;
			*p_long = MEM_IS_VMEM;
		}
	} else {
		UINT32 *p_long = (UINT32*)ptr;
		*p_long = MEM_IS_KMEM;
	}

	{
		UINT8 *p_char = (UINT8*)ptr;
		p_char += sizeof(UINT32);
		*p = p_char;
	}

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

/*! Free memory that has been allocated by osif_malloc. \a ptr points to the
  pointer returned by osif_malloc().
*/
OSIF_CALLTYPE INT32 osif_free(VOID *ptr) {
	UINT32 *free_ptr = (UINT32*)((UINT8*)ptr - sizeof(UINT32));

	if (ptr) {
		if (*free_ptr == MEM_IS_KMEM) {
			kfree(free_ptr);
		} else {
			OSIF_DBG((OSIF_ZONE_OSIF, "%s: using vfree\n", OSIF_FUNCTION));
			vfree(free_ptr);
		}
	}
	return OSIF_SUCCESS;
}

/*! Initialize a memory region pointed by \a ptr to a constant value \a val.
  \a size is the size of the region.
*/
OSIF_CALLTYPE INT32 osif_memset(VOID *ptr, UINT32 val, UINT32 size)
{
	if (ptr) {
		memset(ptr, (char)val, size);
	}
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_memcpy(VOID *dst, VOID *src, UINT32 size)
{
	memcpy(dst, src, size);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_memcmp(VOID *dst, VOID *src, UINT32 size)
{
	return memcmp(dst, src, size);
}

OSIF_CALLTYPE INT32 osif_pci_read_config_byte(OSIF_PCIDEV pcidev, INT32 offs, UINT8 *ptr)
{
	pci_read_config_byte(pcidev, offs, ptr);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_read_config_word(OSIF_PCIDEV pcidev, INT32 offs, UINT16 *ptr)
{
	pci_read_config_word(pcidev, offs, ptr);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_read_config_long(OSIF_PCIDEV pcidev, INT32 offs, UINT32 *ptr)
{
	pci_read_config_dword(pcidev, offs, (u32 *)ptr);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_write_config_byte(OSIF_PCIDEV pcidev, INT32 offs, UINT8 val)
{
	pci_write_config_byte(pcidev, offs, val);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_write_config_word(OSIF_PCIDEV pcidev, INT32 offs, UINT16 val)
{
	pci_write_config_word(pcidev, offs, val);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_write_config_long(OSIF_PCIDEV pcidev, INT32 offs, UINT32 val)
{
	pci_write_config_dword(pcidev, offs, val);
	return OSIF_SUCCESS;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
static inline void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
                                         dma_addr_t *dma_handle)
{
        void *virt_ptr;

        virt_ptr = kmalloc(size, GFP_KERNEL);
        *dma_handle = virt_to_bus(virt_ptr);
        return virt_ptr;
}
#define pci_free_consistent(cookie, size, ptr, dma_ptr) kfree(ptr)
#endif

OSIF_CALLTYPE INT32 osif_pci_malloc(OSIF_PCIDEV pcidev, OSIF_PCI_VADDR *vaddr,
                                    OSIF_PCI_PADDR *paddr, OSIF_PCI_MADDR *maddr, UINT32 size)
{
        OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	*paddr =vmalloc( sizeof( **paddr ) );
	if ( NULL == *paddr) {
		return OSIF_INSUFFICIENT_RESOURCES;
	}

	*vaddr = pci_alloc_consistent( pcidev, size, &((*paddr)->sys_dma_addr) );
	if ( NULL == *vaddr ) {
		vfree( *paddr );
		return OSIF_INSUFFICIENT_RESOURCES;
	}

	(*paddr)->addr32 = (UINT32)((*paddr)->sys_dma_addr);
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_pci_free(OSIF_PCIDEV pcidev, OSIF_PCI_VADDR vaddr,
                                  OSIF_PCI_PADDR *paddr, OSIF_PCI_MADDR maddr, UINT32 size)
{
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	pci_free_consistent(pcidev, size, vaddr, (*paddr)->sys_dma_addr );

	vfree( *paddr );

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE UINT32 osif_pci_get_phy_addr(OSIF_PCI_PADDR paddr)
{
        return paddr->addr32;
}

#if ( (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,8)) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)) )
#warning "vsnprintf not implemented in kernels pre 2.4.8"
#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */
/*!
 *  \brief Needed for vsnprintf()
 */
static INT32 skip_atoi(const INT8 **s)
{
	INT32 i = 0;

	while (isdigit(**s))
		i = i*10 + *((*s)++) - '0';
	return i;
}

/*!
 *  \brief Needed for vsnprintf()
 */
static INT8* number(INT8* buf, INT8* end, INT64 num, INT32 base, INT32 size, INT32 precision, INT32 type)
{
	INT8        c, sign, tmp[66];
	const INT8 *digits;
	const INT8  small_digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
	const INT8  large_digits[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	INT32       i;

	digits = (type & LARGE) ? large_digits : small_digits;
	if (type & LEFT)
		type &= ~ZEROPAD;
	if (base < 2 || base > 36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ';
	sign = 0;
	if (type & SIGN) {
		if (num < 0) {
			sign = '-';
			num = -num;
			size--;
		} else if (type & PLUS) {
			sign = '+';
			size--;
		} else if (type & SPACE) {
			sign = ' ';
			size--;
		}
	}
	if (type & SPECIAL) {
		if (base == 16)
			size -= 2;
		else if (base == 8)
			size--;
	}
	i = 0;
	if (num == 0)
		tmp[i++]='0';
	else while (num != 0)
		tmp[i++] = digits[do_div(num,base)];
	if (i > precision)
		precision = i;
	size -= precision;
	if (!(type&(ZEROPAD+LEFT))) {
		while(size-->0) {
			if (buf <= end)
				*buf = ' ';
			++buf;
		}
	}
	if (sign) {
		if (buf <= end)
			*buf = sign;
		++buf;
	}
	if (type & SPECIAL) {
		if (base==8) {
			if (buf <= end)
				*buf = '0';
			++buf;
		} else if (base==16) {
			if (buf <= end)
				*buf = '0';
			++buf;
			if (buf <= end)
				*buf = digits[33];
			++buf;
		}
	}
	if (!(type & LEFT)) {
		while (size-- > 0) {
			if (buf <= end)
				*buf = c;
			++buf;
		}
	}
	while (i < precision--) {
		if (buf <= end)
			*buf = '0';
		++buf;
	}
	while (i-- > 0) {
		if (buf <= end)
			*buf = tmp[i];
		++buf;
	}
	while (size-- > 0) {
		if (buf <= end)
			*buf = ' ';
		++buf;
	}
	return buf;
}

/*!
 * \brief Format a string and place it in a buffer
 *
 * \param buf The buffer to place the result into
 * \param size The size of the buffer, including the trailing null space
 * \param fmt The format string to use
 * \param args Arguments for the format string
 *
 *  This function was copied directly out of the 2.4.18.SuSE-kernel.
 *  Since kernels pre 2.4.6 don't support this function, we need to
 *  provide it on our own.
 */
OSIF_CALLTYPE INT32 osif_vsnprintf(INT8 *buf, size_t size, const INT8 *fmt, va_list args)
{
	INT32       len;
	UINT64      num;
	INT32       i, base;
	INT8       *str, *end, c;
	const INT8 *s;

	INT32 flags;		/* flags to number() */

	INT32 field_width;	/* width of output field */
	INT32 precision;	/* min. # of digits for integers; max
				   number of chars for from string */
	INT32 qualifier;	/* 'h', 'l', or 'L' for integer fields */
				/* 'z' support added 23/7/1999 S.H.    */
				/* 'z' changed to 'Z' --davidm 1/25/99 */

	str = buf;
	end = buf + size - 1;

	if (end < buf - 1) {
		end = ((void *) (-1));
		size = end - buf + 1;
	}

	for (; *fmt ; ++fmt) {
		if (*fmt != '%') {
			if (str <= end)
				*str = *fmt;
			++str;
			continue;
		}

		/* process flags */
		flags = 0;
		repeat:
			++fmt;		/* this also skips first '%' */
			switch (*fmt) {
			case '-': flags |= LEFT; goto repeat;
			case '+': flags |= PLUS; goto repeat;
				case ' ': flags |= SPACE; goto repeat;
				case '#': flags |= SPECIAL; goto repeat;
				case '0': flags |= ZEROPAD; goto repeat;
			}

		/* get field width */
		field_width = -1;
		if (isdigit(*fmt))
			field_width = skip_atoi(&fmt);
		else if (*fmt == '*') {
			++fmt;
			/* it's the next argument */
			field_width = va_arg(args, INT32);
			if (field_width < 0) {
				field_width = -field_width;
				flags |= LEFT;
			}
		}

		/* get the precision */
		precision = -1;
		if (*fmt == '.') {
			++fmt;
			if (isdigit(*fmt))
				precision = skip_atoi(&fmt);
			else if (*fmt == '*') {
				++fmt;
				/* it's the next argument */
				precision = va_arg(args, INT32);
			}
			if (precision < 0)
				precision = 0;
		}

		/* get the conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L' || *fmt =='Z') {
			qualifier = *fmt;
			++fmt;
			if (qualifier == 'l' && *fmt == 'l') {
				qualifier = 'L';
				++fmt;
			}
		}

		/* default base */
		base = 10;

		switch (*fmt) {
			case 'c':
				if (!(flags & LEFT)) {
					while (--field_width > 0) {
						if (str <= end)
							*str = ' ';
						++str;
					}
				}
				c = (UINT8) va_arg(args, INT32);
				if (str <= end)
					*str = c;
				++str;
				while (--field_width > 0) {
					if (str <= end)
						*str = ' ';
					++str;
				}
				continue;

			case 's':
				s = va_arg(args, INT8*);
				if (!s)
					s = "<NULL>";

				len = strnlen(s, precision);

				if (!(flags & LEFT)) {
					while (len < field_width--) {
						if (str <= end)
							*str = ' ';
						++str;
					}
				}
				for (i = 0; i < len; ++i) {
					if (str <= end)
						*str = *s;
					++str; ++s;
				}
				while (len < field_width--) {
					if (str <= end)
						*str = ' ';
					++str;
				}
				continue;

			case 'p':
				if (field_width == -1) {
					field_width = 2*sizeof(void*);
					flags |= ZEROPAD;
				}
				str = number(str, end,
						(INT32)va_arg(args, void*),
						16, field_width, precision, flags);
				continue;


			case 'n':
				/* FIXME:
				* What does C99 say about the overflow case here? */
				if (qualifier == 'l') {
					INT32 *ip = va_arg(args, INT32*);
					*ip = (str - buf);
				} else if (qualifier == 'Z') {
					size_t *ip = va_arg(args, size_t*);
					*ip = (str - buf);
				} else {
					INT32 *ip = va_arg(args, INT32*);
					*ip = (str - buf);
				}
				continue;

			case '%':
				if (str <= end)
					*str = '%';
				++str;
				continue;

				/* integer number formats - set up the flags and "break" */
			case 'o':
				base = 8;
				break;

			case 'X':
				flags |= LARGE;
			case 'x':
				base = 16;
				break;

			case 'd':
			case 'i':
				flags |= SIGN;
			case 'u':
				break;

			default:
				if (str <= end)
					*str = '%';
				++str;
				if (*fmt) {
					if (str <= end)
						*str = *fmt;
					++str;
				} else {
					--fmt;
				}
				continue;
		}
		if (qualifier == 'L')
			num = va_arg(args, INT64);
		else if (qualifier == 'l') {
			num = va_arg(args, UINT32);
			if (flags & SIGN)
				num = (INT32) num;
		} else if (qualifier == 'Z') {
			num = va_arg(args, size_t);
		} else if (qualifier == 'h') {
			num = (UINT16)va_arg(args, INT32);
			if (flags & SIGN)
				num = (INT16) num;
		} else {
			num = va_arg(args, UINT32);
			if (flags & SIGN)
				num = (INT32) num;
		}
		str = number(str, end, num, base,
				field_width, precision, flags);
	}
	if (str <= end)
		*str = '\0';
	else if (size > 0)
		/* don't write out a null byte if the buf size is zero */
		*end = '\0';
	/* the trailing null byte doesn't count towards the total
	* ++str;
	*/
	return str-buf;
}
#endif

/*! This function prints a formated text string.
*/
OSIF_CALLTYPE INT32 osif_print(const INT8 *fmt, ...)
{
	va_list	arglist;
	INT8    output[400];
	INT32   len;

	va_start(arglist, fmt);
	len = OSIF_VSNPRINTF(output, 400, fmt, arglist);
	va_end(arglist);
	printk(output);
	return len;
}

/*! This function prints a formated text string, only, if one of the bits
 *  specified in output_mask was set at module-load-time (global variable verbose).
 */
OSIF_CALLTYPE INT32 osif_dprint(UINT32 output_mask, const INT8 *fmt, ...)
{
	va_list	arglist;
	INT8    output[400];
	INT32   len = 0;

#ifndef ESDDBG
	output_mask &= 0x000000FF;
#endif
	if ( verbose & output_mask ) {
		va_start(arglist, fmt);
		len = OSIF_VSNPRINTF(output, 400, fmt, arglist);
		va_end(arglist);
		printk(output);
	}
	return len;
}

OSIF_CALLTYPE INT32 osif_snprintf(UINT8 *str, INT32 size, const INT8 *format, ...)
{
	va_list	arglist;
	INT32   len;

	va_start(arglist, format);
        len = OSIF_VSNPRINTF(str, size, format, arglist);
	va_end(arglist);
	return len;
}

/*! Sleep for a given number (\a ms) of milli seconds.
*/
OSIF_CALLTYPE INT32 osif_sleep( INT32 ms )
{
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        wait_queue_head_t wq;
        INT32  result = OSIF_SUCCESS;
        UINT32 timeout = ((ms*HZ)+999)/1000;

        init_waitqueue_head(&wq);
        result = wait_event_interruptible_timeout(wq, (timeout == 0), timeout);
        if (result == -ERESTARTSYS) {
                OSIF_DBG((OSIF_ZONE_OSIF, "%s: Sleep cancelled by SIGTERM!", OSIF_FUNCTION));
        }
#else
	OSIF_DECL_WAIT_QUEUE(p);
	interruptible_sleep_on_timeout( &p, ((ms*HZ)+999)/1000);
#endif
	return OSIF_SUCCESS;
}

/*! Initialize the semaphore pointed by \a s to a given value.
*/
OSIF_CALLTYPE INT32 osif_sema_create(OSIF_SEMA *s, INT32 init)
{
	*s = vmalloc(sizeof(**s));
	if(NULL == *s){
		return OSIF_INSUFFICIENT_RESOURCES;
	}
#if 1 /* MT */
	memset(*s, 0x0, sizeof(**s));
#endif
	sema_init(&((*s)->sema), init);
	return OSIF_SUCCESS;
}

/*! Initialize the semophore pointed by \a s to a given value.
*/
OSIF_CALLTYPE INT32 osif_sema_destroy(OSIF_SEMA *s)
{
	vfree(*s);
	*s = NULL;
	return OSIF_SUCCESS;
}

/*! Decrease a semaphore's \a s value or sleep on it, when
  the value is zero.
*/
OSIF_CALLTYPE INT32 osif_sema_down(OSIF_SEMA *s)
{
    	down((&((*s)->sema)));
    	return OSIF_SUCCESS;
}

/*! Decrease a semaphore's \a s value or sleep on it, when
  the value is zero (interruptible).
*/
OSIF_CALLTYPE INT32 osif_sema_down_interruptible(OSIF_SEMA *s)
{
	if (down_interruptible(&((*s)->sema)) < 0) {
		return OSIF_OPERATION_ABORTED;
	}
	return OSIF_SUCCESS;
}

/*! Increase a semaphore's \a s value and wake up processes that are
  waiting for this semaphore.
*/
OSIF_CALLTYPE INT32 osif_sema_up(OSIF_SEMA *s)
{
	up(&((*s)->sema));
	return OSIF_SUCCESS;
}

/*
** mutex entries
*/
OSIF_CALLTYPE INT32 osif_mutex_create(OSIF_MUTEX *m)
{
	*m = vmalloc(sizeof(**m));
	if(NULL == *m) {
		return OSIF_INSUFFICIENT_RESOURCES;
	}
#if 1 /* MT */
	memset(*m, 0x00, sizeof(**m));
#endif
	sema_init(&((*m)->sema), 1);
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_mutex_destroy(OSIF_MUTEX *m)
{
	vfree(*m);
	*m=NULL;
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_mutex_lock(OSIF_MUTEX *m)
{
	down(&((*m)->sema));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_mutex_unlock(OSIF_MUTEX *m)
{
	up(&((*m)->sema));
	return OSIF_SUCCESS;
}

/*
 * mutexes for short sequences which do not call the kernel!
 */
OSIF_CALLTYPE INT32 osif_busy_mutex_create(OSIF_BUSY_MUTEX *m)
{
	*m = vmalloc(sizeof(**m));
	if(NULL == *m) {
		return OSIF_INSUFFICIENT_RESOURCES;
	}

	(*m)->spinlock =SPIN_LOCK_UNLOCKED;
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_busy_mutex_destroy(OSIF_BUSY_MUTEX *m)
{
	vfree(*m);
	*m=NULL;
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_busy_mutex_lock(OSIF_BUSY_MUTEX *m)
{
	spin_lock(&((*m)->spinlock));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_busy_mutex_unlock(OSIF_BUSY_MUTEX *m)
{
	spin_unlock(&((*m)->spinlock));
	return OSIF_SUCCESS;
}

/*
 * mutexes for synchronisation against the irq-level
 */
OSIF_CALLTYPE INT32 osif_irq_mutex_create(OSIF_IRQ_MUTEX *im)
{
	*im =vmalloc(sizeof(**im));
	if(NULL == *im) {
		return OSIF_INSUFFICIENT_RESOURCES;
	}

	(*im)->spinlock =SPIN_LOCK_UNLOCKED;
#ifdef OSIF_DEBUG
	(*im)->magic = 0x0;
#endif
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_irq_mutex_destroy(OSIF_IRQ_MUTEX *im)
{
	vfree(*im);
	*im=NULL;
	return OSIF_SUCCESS;
}

/*
 *  Call this entry, if you are not inside in an irq handler
 */
OSIF_CALLTYPE INT32 osif_irq_mutex_lock(OSIF_IRQ_MUTEX *im)
{
#ifdef OSIF_DEBUG
	struct task_struct *tsk = current;
#endif
	spin_lock_irq(&((*im)->spinlock));
#ifdef OSIF_DEBUG
	(*im)->magic = tsk->pid | 0x80000000; /* MSB=1 marks entry */
#endif
	return OSIF_SUCCESS;
}

/*
 *  Call this entry, if you are not inside an irq handler
 */
OSIF_CALLTYPE INT32 osif_irq_mutex_unlock(OSIF_IRQ_MUTEX *im)
{
#ifdef OSIF_DEBUG
	struct task_struct *tsk = current;
	(*im)->magic = tsk->pid; /* MSB=0 marks exit (assuming pid's smaller than 0x7FFFFFFF) */
#endif
	spin_unlock_irq(&((*im)->spinlock));
	return OSIF_SUCCESS;
}

/*
 *  Call this entry inside irq handlers
 */
OSIF_CALLTYPE INT32 osif_in_irq_mutex_lock(OSIF_IRQ_MUTEX *im)
{
	spin_lock_irq(&((*im)->spinlock));
#ifdef OSIF_DEBUG
	(*im)->magic = 0xABABABAB;
#endif
	return OSIF_SUCCESS;
}

/*
 *  Call this entry inside irq handlers
 */
OSIF_CALLTYPE INT32 osif_in_irq_mutex_unlock(OSIF_IRQ_MUTEX *im)
{
#ifdef OSIF_DEBUG
	(*im)->magic = 0x2BABABAB;
#endif
	spin_unlock_irq(&((*im)->spinlock));
	return OSIF_SUCCESS;
}

static int osif_dpc_thread( void *context )
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct task_struct *tsk = current;
#endif
	OSIF_DPC dpc;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));
        daemonize("dpc thread");
/* AB TODO: set affinity to CPU, which executes interrupts (to get accurate timestamps...)
long sched_setaffinity(pid_t pid, cpumask_t new_mask)
long sched_getaffinity(pid_t pid, cpumask_t *mask)
 */

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter tsk=%p\n", OSIF_FUNCTION, tsk));
        daemonize();
        /* todo: do we need this ? */ /*  tsk->pgrp = 1; */
        sigfillset(&tsk->blocked); /* block all signals */
        strcpy(tsk->comm,"dpc thread");
#else
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter tsk=%p\n", OSIF_FUNCTION, tsk));
        tsk->session = 1;
        tsk->pgrp = 1;
        strcpy(tsk->comm, "dpc thread");
        spin_lock_irq(&tsk->sigmask_lock);
        sigfillset(&tsk->blocked);             /* block all signals */
        recalc_sigpending(tsk);
        spin_unlock_irq(&tsk->sigmask_lock);
#endif

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: can dpc running\n", OSIF_FUNCTION));

       /* notify that backend thread is running */
	dpc_root.thread_running = 1;
	up( &dpc_root.sema_notify );

 	while (!dpc_root.destroy) {
#if 0    /* removed: because all signals are blocked here !?*/
	       	down_interruptible(&dpc_root.sema_trigger);
#else
		down(&dpc_root.sema_trigger);
#endif
		if (dpc_root.destroy) {
			break;
		}

		/* call dpc function */
		OSIF_DBG((OSIF_ZONE_OSIF, "%s: dpc triggered\n", OSIF_FUNCTION));

		spin_lock_irq( &dpc_root.spinlock );
		while( NULL != dpc_root.first ) {
			dpc = dpc_root.first;
			dpc_root.first = dpc->next;    /* link out          */
			dpc->linked = 0;               /* mark: linked out  */
			dpc->running = 1;              /* mark: running     */
			spin_unlock_irq( &dpc_root.spinlock );
			dpc->func(dpc->arg);
			spin_lock_irq( &dpc_root.spinlock );
			dpc->running = 0;              /* mark: not running */
		}

		spin_unlock_irq( &dpc_root.spinlock );
	}

	/* notify that backend thread is exiting */
	dpc_root.thread_running = 0;
	up( &dpc_root.sema_notify );

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	complete_and_exit(&dpcThreadCompletion, 0);
#endif
	return 0;  /* This will never reached */
}

static INT32 osif_dpc_root_create( VOID)
{
        INT32 threadId;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	dpc_root.thread_running = 0;

	dpc_root.spinlock = SPIN_LOCK_UNLOCKED;
	sema_init( &dpc_root.sema_trigger, 0 );
	sema_init( &dpc_root.sema_notify, 0 );

	threadId = kernel_thread(osif_dpc_thread, 0, 0);
        if ( 0 == threadId ) {
                return CANIO_EIO;
        }

	down( &dpc_root.sema_notify );

	if ( 0 == dpc_root.thread_running ) {
		return OSIF_INSUFFICIENT_RESOURCES;
	}

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}


static INT32 osif_dpc_root_destroy( VOID)
{
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	dpc_root.destroy = 1;
	up( &dpc_root.sema_trigger );
	down(&dpc_root.sema_notify );
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	wait_for_completion(&dpcThreadCompletion);
#endif
	dpc_root.destroy = 0;
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_dpc_create( OSIF_DPC *dpc, OSIF_CALLTYPE VOID(*func)(VOID *), VOID *arg )
{
	INT32 result;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	result = OSIF_MALLOC( sizeof(**dpc), (VOID*)dpc );
	if ( OSIF_SUCCESS != result ) {
		return result;
	}

	OSIF_MEMSET( *dpc, 0, sizeof(**dpc) );

	(*dpc)->func = func;
	(*dpc)->arg  = arg;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_dpc_destroy( OSIF_DPC *dpc )
{
	OSIF_DPC  dpc2;
	UINTPTR   flags;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	if (NULL == *dpc) {
		return CANIO_EFAULT;
	}

	spin_lock_irqsave( &dpc_root.spinlock, flags );
	(*dpc)->destroy = 1;                       /* now triggers will ret ENOENT */

	if(*dpc == dpc_root.first) {
		dpc_root.first = (*dpc)->next;     /* link out         */
		(*dpc)->next = NULL;               /* mark: linked out */
	} else {
		dpc2 = dpc_root.first;
		while( NULL != dpc2 ) {
			if(*dpc == dpc2->next) {
				dpc2->next = (*dpc)->next;         /* link out */
				(*dpc)->next = NULL;               /* mark: linked out   */
				break;
			}

			dpc2 = dpc2->next;                      /* move up */
		}
	}

	/* if it was linked, it is now unlinked and ready for removal */
	/* if it was unlinked, it could just run... */

	while( 0 != (*dpc)->running ) {
		spin_unlock_irqrestore( &dpc_root.spinlock, flags );
		OSIF_SLEEP(300);
		spin_lock_irqsave( &dpc_root.spinlock, flags );
	}

	spin_unlock_irqrestore( &dpc_root.spinlock, flags );

	OSIF_FREE(*dpc);
	*dpc = NULL;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

/*
 * DO NOT RUN THIS OUTSIDE IRQ-LEVEL!
 */
OSIF_CALLTYPE INT32 osif_dpc_trigger( OSIF_DPC *dpc )
{
	UINTPTR flags;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter, dpc=%p\n", OSIF_FUNCTION,(UINTPTR)dpc ));

	if (NULL == *dpc) {
		return CANIO_EFAULT;
	}

	spin_lock_irqsave( &dpc_root.spinlock, flags );

	if( 0 != (*dpc)->destroy ) {
		spin_unlock_irqrestore( &dpc_root.spinlock, flags );
		return CANIO_ENOENT;
	}

	if( 0 == (*dpc)->linked ) {
		if( NULL == dpc_root.first) {
			dpc_root.first = *dpc;
		} else {
			OSIF_DPC dpc2;

			dpc2 = dpc_root.first;
			while( NULL != dpc2->next ) {
				dpc2 = dpc2->next;
			}

			dpc2->next = *dpc;
		}

		(*dpc)->next = NULL;     /* mark end of chain */
		(*dpc)->linked = 1;      /* mark linked */

		spin_unlock_irqrestore( &dpc_root.spinlock, flags );
		up(&dpc_root.sema_trigger);
	} else {
		spin_unlock_irqrestore( &dpc_root.spinlock, flags );
	}

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

/*
** this timer stub runs on soft-irq-level and triggers a dpc for the user timer handler
*/
static void timer_stub( _OSIF_TIMER *timer)
{
	OSIF_DPC_TRIGGER( &timer->dpc );
}


/*! \fn osif_timer_create( OSIF_TIMER **timer, VOID(*func)(VOID *), VOID *arg );
 *  \brief Create a timer object.
 *  \param timer a pointer to the created timer structure is returned through this parameter.
 *  \param func func(arg) will be called when timer expires.
 *  \param arg is passed as argument to the timer handler \a func.
 *  The created timer is initially disabled after creation.
 */
OSIF_CALLTYPE INT32 osif_timer_create( OSIF_TIMER *t, VOID (OSIF_CALLTYPE *func)(VOID *), VOID *arg )
{
	INT32 result;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	result = OSIF_MALLOC(sizeof(**t), t);
	if (result != OSIF_SUCCESS) {
		return result;
	}

	init_timer(&((*t)->timer));
	(*t)->timer.function = (VOID *)timer_stub;
	(*t)->timer.data     = (UINTPTR)(*t);
	(*t)->timer.expires  = 0;

	OSIF_DPC_CREATE(&(*t)->dpc, func, arg);

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

/*! \fn osif_timer_destroy( OSIF_TIMER *timer );
 *  \brief Destroy a timer object.
 *  \param timer a pointer to the timer structure.
 */
OSIF_CALLTYPE INT32 osif_timer_destroy( OSIF_TIMER *t )
{
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	if (NULL == *t) {
		return CANIO_ERROR;
	}

	del_timer(&((*t)->timer));
	OSIF_DPC_DESTROY(&(*t)->dpc);

	OSIF_FREE(*t);
	*t=NULL;
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return CANIO_ERROR;
}

/*! \fn osif_timer_set( OSIF_TIMER *timer, UINT32 ms );
 *  \brief Set the expiration time for a timer.
 *  \param timer is a pointer to the timer that should be modified
 *  \param ms new expiration time for this timer in ms. A value of 0
 *  disables the timer.
 */
OSIF_CALLTYPE INT32 osif_timer_set( OSIF_TIMER *t, UINT32 ms )
{
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter (ms = %d)\n", OSIF_FUNCTION, ms));

	if (ms) {
		mod_timer(&((*t)->timer), jiffies + (((ms * HZ) + 500) / 1000)); /* round correctly */
	} else {
		del_timer(&((*t)->timer));
	}

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave\n", OSIF_FUNCTION));
	return OSIF_SUCCESS;
}

OSIF_CALLTYPE INT32 osif_timer_get( OSIF_TIMER *t, UINT32 *ms )
{
	OSIF_DBG((OSIF_ZONE_OSIF, "%s: enter\n", OSIF_FUNCTION));

	if (ms)
		*ms = jiffies * 1000 / HZ;

	OSIF_DBG((OSIF_ZONE_OSIF, "%s: leave (ms=%d)\n", OSIF_FUNCTION, *ms));
	return OSIF_SUCCESS;
}

#ifdef OSIF_ARCH_PPC
UINT64 __attribute__((__unused__)) osif_ticks( VOID )
{
	unsigned __lo;
	unsigned __hi;
	unsigned __tmp;

	__asm__ __volatile__(
		"1:;"
		"  mftbu	%0;"
		"  mftb		%1;"
		"  mftbu	%2;"
		"  cmplw	%0,%2;"
		"bne- 1b;"
		: "=r" (__hi), "=r" (__lo), "=r" (__tmp)
	);
	return (UINT64) __hi << 32 | (UINT64) __lo;
}
#endif

#ifdef OSIF_ARCH_X86
OSIF_CALLTYPE UINT64 osif_ticks( VOID )
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return ((UINT64)tv.tv_sec * 1000000LL) + (UINT64)tv.tv_usec;
}
#endif

#ifdef OSIF_ARCH_PPC
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
extern bd_t __res;
#endif

OSIF_CALLTYPE UINT64 osif_tick_frequency()
{
#if !(defined(BOARD_cpci750)) && !(defined(BOARD_mecp52))
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	bd_t *bd = (bd_t *)&__res;
# else
	bd_t *bd = (bd_t *)__res; /* we find 'unsigned char __res[]' in old kernels */
# endif
 	OSIF_DBG((OSIF_ZONE_OSIF, "%s: cpu freq = %d\n", OSIF_FUNCTION, bd->bi_procfreq));
	return (UINT64)(bd->bi_procfreq);
#else
	/* TODO: get CPU frequency from somewhere in kernel */
	return (UINT64)(33333000);
#endif
}
#endif

OSIF_CALLTYPE UINT32 osif_div64_32(UINT64 *n, UINT32 base)
{
	UINT64 rem = *n;
	UINT64 b = base;
	UINT64 res, d = 1;
	UINT32 high = rem >> 32;

	/* Reduce the thing a bit first */
	res = 0;
	if (high >= base) {
		high /= base;
		res = (UINT64) high << 32;
		rem -= (UINT64) (high*base) << 32;
	}

	while ((INT64)b > 0 && b < rem) {
		b <<= 1;
		d <<= 1;
	}

	do {
		if (rem >= b) {
			rem -= b;
			res += d;
		}
		b >>= 1;
		d >>= 1;
	} while (d);

	*n = res;
	return rem;
}

OSIF_CALLTYPE VOID osif_ser_out( UINT8 *str )
{
        UINT32 i;

        /* disable serial interrupt */
        outb( (unsigned char)0, 0x3F9 );

        /* output of string */
        for (i = 0; i < strlen(str); i++) {
                while ( !(inb(0x3F8+5) & 0x20) );
                outb( str[i], 0x3F8 );
        }
        while ( !(inb(0x3F8+5) & 0x20) );

        /* reenable serial interrupt */
/*        outb(0x3F9, 0x??); */ /* AB TODO */

        return;
}

OSIF_CALLTYPE INT32 osif_load_from_file( UINT8 *destaddr, UINT8 *filename )
{
	struct file    *srcf;
	int            retval,orgfsuid,orgfsgid;
	mm_segment_t   orgfs;
	char           *buffer;
	unsigned long  page;
	int            result = OSIF_SUCCESS;

	/* Save uid and gid used for filesystem access.
	 * Set user a/home/andreas/tmpinitrdnd group to 0 (root)
	 */
	orgfsuid=current->fsuid;
	orgfsgid=current->fsgid;
	current->fsuid=current->fsgid=0;
	/* save FS register and set FS register to kernel
	 * space, needed for read and write to accept
	 * buffer in kernel space.
	 */
	orgfs=get_fs();
	set_fs(KERNEL_DS);

	if ( filename && *filename ) {
		/* Allocate one page for buffer */
		page = __get_free_page(GFP_KERNEL);
		if (page) {
			buffer=(char*)page;
			srcf = filp_open(filename, O_RDONLY, 0);
			if (IS_ERR(srcf)) {
				OSIF_PRINT(("ERROR: Cannot open file <%s>.\n",filename));
				result = OSIF_INVALID_PARAMETER;
			} else {
				/* The object must have a read method */
				if (srcf->f_op&&srcf->f_op->read) {
					do {
						/* Read as much as posible into the buffer,
						 * at most one page.
						 */
						retval=srcf->f_op->read(srcf,buffer,PAGE_SIZE,&srcf->f_pos);
						if ( retval < 0 ) {
							OSIF_DBG((OSIF_ZONE_OSIF, "%s: Read error %d\n", OSIF_FUNCTION,-retval));
						}
						if (retval>0) {
							int i;
							/* Continue writing until error or everything
							 * written.
							 */
							for (i=0; i< retval; i++) {
								*(volatile UINT8*)destaddr++ = (volatile UINT8)buffer[i];
							}
						}
					} while (retval>0);
					/* Now clean up everything that was actually allocated */
				} else {
					OSIF_DBG((OSIF_ZONE_OSIF, "%s: %s does not have a read method\n", OSIF_FUNCTION,filename));
					result = CANIO_ERROR;
				}
				retval=filp_close(srcf,NULL);
				if (retval) {
					OSIF_DBG((OSIF_ZONE_OSIF, "%s: Error %d closing %s\n", OSIF_FUNCTION,-retval,filename));
					result = CANIO_ERROR;
				}
			}
			free_page(page);
		} else {
			OSIF_DBG((OSIF_ZONE_OSIF, "%s: Out of memory\n", OSIF_FUNCTION));
			result = OSIF_INSUFFICIENT_RESOURCES;
		}
	} else {
		OSIF_DBG((OSIF_ZONE_OSIF, "%s: Invalid filename\n", OSIF_FUNCTION));
		result = OSIF_INVALID_PARAMETER;
	}

	set_fs(orgfs);
	current->fsuid=orgfsuid;
	current->fsgid=orgfsgid;

	OSIF_DBG((OSIF_ZONE_OSIF,  "%s: leave, result=%08x\n", OSIF_FUNCTION, result ));
	return result;
}

/*! \fn osif_io_out8(void *addr, UINT8 data);
 *  \brief Write access to I/O-space (8 bit wide)
 *  \param addr target address
 *  \param data this data is written to target address
 */
OSIF_CALLTYPE void osif_io_out8(UINT8 *addr, UINT8 data)
{
        outb((UINT8)data, (UINTPTR)(addr));
}

/*! \fn osif_io_in8(void *addr);
 *  \brief Read access to I/O-space (8 bit wide)
 *  \param addr target address
 *  \return data from target address
 */
OSIF_CALLTYPE UINT8 osif_io_in8(UINT8 *addr)
{
        return inb((UINTPTR)addr);
}

/*! \fn osif_io_out16(void *addr, UINT16 data);
 *  \brief Write access to I/O-space (16 bit wide)
 *  \param addr target address
 *  \param data this data is written to target address
 */
OSIF_CALLTYPE void osif_io_out16(UINT16 *addr, UINT16 data)
{
        outw((UINT16)data, (UINTPTR)(addr));
}

/*! \fn osif_io_in16(void *addr);
 *  \brief Read access to I/O-space (16 bit wide)
 *  \param addr target address
 *  \return data from target address
 */
OSIF_CALLTYPE UINT16 osif_io_in16(UINT16 *addr)
{
        return inw((UINTPTR)addr);
}

/*! \fn osif_io_out32(void *addr, UINT32 data);
 *  \brief Write access to I/O-space (32 bit wide)
 *  \param addr target address
 *  \param data this data is written to target address
 */
OSIF_CALLTYPE void osif_io_out32(UINT32 *addr, UINT32 data)
{
        outl((UINT32)data, (UINTPTR)(addr));
}

/*! \fn osif_io_in32(void *addr);
 *  \brief Read access to I/O-space (32 bit wide)
 *  \param addr target address
 *  \return data from target address
 */
OSIF_CALLTYPE UINT32 osif_io_in32(UINT32 *addr)
{
        return inl((UINTPTR)addr);
}

/************************************************************************/
/*                     UTILITY CODE                                     */
/************************************************************************/
OSIF_CALLTYPE INT32 osif_make_ntcan_serial(const INT8* pszSerial,
					   UINT32 *pulSerial)
{
        INT32 lTemp;
        *pulSerial = 0;

        if(pszSerial[0] >= 'A' && pszSerial[0] <= 'P' &&
           pszSerial[1] >= 'A' && pszSerial[1] <= 'P')	{
                lTemp = ((pszSerial[0] - 'A') << 28) | ((pszSerial[1] - 'A') << 24);
                *pulSerial = simple_strtoul(&pszSerial[2], NULL, 10);
                *pulSerial |= (UINT32)lTemp;
        }

        return OSIF_SUCCESS;
}

/*
 * RCS/CVS id and build tag with support to prevent compiler warnings about
 * unused vars for different compiler
 */
#if defined(_MSC_VER) && (_MSC_VER >= 800)                           /* VC++ 32 Bit compiler */
 static char* rcsid = "$Id: osif.c,v 1.80 2006/10/12 13:31:55 manuel Exp $";
#elif defined (__ghs__)                         /* Green Hills compiler */
# pragma ident "$Id: osif.c,v 1.80 2006/10/12 13:31:55 manuel Exp $"
#elif ((__GNUC__ > 2) || (__GNUC__ == 2 && __GNUC_MINOR__ >=7)) /* GCC > 2.7 */
 static char* rcsid __attribute__((unused)) = "$Id: osif.c,v 1.80 2006/10/12 13:31:55 manuel Exp $";
#else                                           /* No or old GNU compiler */
# define USE(var) static void use_##var(void *x) {if(x) use_##var((void *)var);}
 static char* rcsid = "$Id: osif.c,v 1.80 2006/10/12 13:31:55 manuel Exp $";
 USE(rcsid);
#endif /* of _MSC_VER >= 800 */
