/* -*- linux-c -*-
 * FILE NAME osif.h
 *           copyright 2002-2006 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *           ...
 *
 *
 *
 * history:
 *
 *  24.05.02 - first version                                        mf
 *  27.10.03 - New output function osif_dprint (with verbose mask)  ab
 *  13.11.03 - OSIF_BUSY_MUTEX_xxx added                            mt
 *  11.05.04 - Removed OSIF_EXTERN, added OSIF_CALLTYPE             ab
 *  23.09.04 - Corrected endianess-macros and memcmp (please define
 *             them in the osifi.h the respective system)           ab
 *  28.09.06 - Added OSIF_PRIx64, OSIF_PRId64, OSIF_PRIu64          mk
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2006 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file osif.h
    \brief OSIF header

    This file contains the OSIF prototypes and user macros.
    Any driver should call the OSIF_<xxx> macros instead of the
    osif_<xxx> functions.
*/

#ifndef __OSIF_H__
#define __OSIF_H__


#define OSIF_KERNEL

#define OSIF_CALLTYPE
#define OSIF_INLINE   inline
#define OSIF_STATIC   static
#define OSIF_REGISTER register

#define OSIF_FUNCTION __FUNCTION__

#include <osiftypes.h>
#include <osifi.h>
#ifdef OSIF_MEMDBG
# include <memdbg.h>
#endif
#define OSIF_ATTACH()          osif_attach()
#define OSIF_DETACH()          osif_detach()

#ifdef OSIF_MEMDBG
# define OSIF_MALLOC(s,p)      osif_memdbg_malloc((s),(VOID**)(p), __FILE__, __LINE__)
# define OSIF_FREE(p)          osif_memdbg_free((p))
#else
# define OSIF_MALLOC(s,p)      osif_malloc((s),(VOID**)(p))      /*!< see osif_malloc(...) */
# define OSIF_FREE(p)          osif_free((p))            /*!< see osif_free(...) */
#endif
#define OSIF_MEMSET(p,v,s)     osif_memset((p),(v),(s))  /*!< see osif_memset(...) */
#define OSIF_MEMCPY(d,s,n)     osif_memcpy((d),(s),(n))
#define OSIF_MEMCMP(d,s,n)     osif_memcmp((d),(s),(n))

#define	OSIF_SEMA_CREATE(s,i)  osif_sema_create((s),(i))
#define OSIF_SEMA_DESTROY(s)   osif_sema_destroy(s)
#define OSIF_SEMA_DOWN(s)      osif_sema_down(s)
#define OSIF_SEMA_DOWN_INTERRUPTIBLE(s) osif_sema_down_interruptible(s)
#define OSIF_SEMA_UP(s)        osif_sema_up(s)

#define	OSIF_MUTEX_CREATE(m)   osif_mutex_create(m)
#define OSIF_MUTEX_DESTROY(m)  osif_mutex_destroy(m)
#define OSIF_MUTEX_LOCK(m)     osif_mutex_lock(m)
#define OSIF_MUTEX_UNLOCK(m)   osif_mutex_unlock(m)

#define OSIF_BUSY_MUTEX_CREATE(m)   osif_busy_mutex_create(m)
#define OSIF_BUSY_MUTEX_DESTROY(m)  osif_busy_mutex_destroy(m)
#define OSIF_BUSY_MUTEX_LOCK(m)     osif_busy_mutex_lock(m)
#define OSIF_BUSY_MUTEX_UNLOCK(m)   osif_busy_mutex_unlock(m)

#define OSIF_IRQ_MUTEX_CREATE(im)    osif_irq_mutex_create(im)
#define OSIF_IRQ_MUTEX_DESTROY(im)   osif_irq_mutex_destroy(im)
#define OSIF_IRQ_MUTEX_LOCK(im)      osif_irq_mutex_lock(im)
#define OSIF_IRQ_MUTEX_UNLOCK(im)    osif_irq_mutex_unlock(im)
#define OSIF_IN_IRQ_MUTEX_LOCK(im)   osif_in_irq_mutex_lock(im)
#define OSIF_IN_IRQ_MUTEX_UNLOCK(im) osif_in_irq_mutex_unlock(im)

#define OSIF_PRINT(fmt)               osif_print fmt
#define OSIF_DPRINT(fmt)              osif_dprint fmt
#define OSIF_SNPRINTF(args)           osif_snprintf args
#define OSIF_SLEEP(ms)                osif_sleep(ms)

#define OSIF_DPC_CREATE(d,f,a) osif_dpc_create((d),(f),(a));
#define OSIF_DPC_DESTROY(d)    osif_dpc_destroy(d);
#define OSIF_DPC_TRIGGER(d)    osif_dpc_trigger(d);

#define OSIF_TIMER_CREATE(t,f,a) osif_timer_create((t),(f),(a));
#define OSIF_TIMER_DESTROY(t)  osif_timer_destroy(t);
#define OSIF_TIMER_SET(t,m)    osif_timer_set((t),(m));
#define OSIF_TIMER_FREQ        1000
#define OSIF_TIMER_GET(t,m)    osif_timer_get((t),(m));

#define OSIF_PCI_MALLOC(d,v,p,m,s) osif_pci_malloc((d),(v),(p),(m),(s));
#define OSIF_PCI_FREE(d,v,p,m,s)   osif_pci_free((d),(v),(p),(m),(s));
#define OSIF_PCI_GET_PHY_ADDR(p)   osif_pci_get_phy_addr(p)
#define	OSIF_PCI_READ_CONFIG_BYTE(a,b,c) osif_pci_read_config_byte((a),(b),(c));
#define	OSIF_PCI_READ_CONFIG_WORD(a,b,c) osif_pci_read_config_word((a),(b),(c));
#define	OSIF_PCI_READ_CONFIG_LONG(a,b,c) osif_pci_read_config_long((a),(b),(c));
#define	OSIF_PCI_WRITE_CONFIG_BYTE(a,b,c) osif_pci_write_config_byte((a),(b),(c));
#define	OSIF_PCI_WRITE_CONFIG_WORD(a,b,c) osif_pci_write_config_word((a),(b),(c));
#define	OSIF_PCI_WRITE_CONFIG_LONG(a,b,c) osif_pci_write_config_long((a),(b),(c));

#define OSIF_IRQ_ATTACH(i,h,n,a) osif_irq_attach((i),(h),(n),(a));
#define OSIF_IRQ_DETACH(i,a) osif_irq_detach((i),(a));

#define OSIF_TICKS             osif_ticks()

#ifndef OSIF_TICK_FREQ
# define OSIF_TICK_FREQ        osif_tick_frequency()
#endif

#define OSIF_IO_OUT8( addr, data ) osif_io_out8( (addr), (data) )
#define OSIF_IO_IN8( addr ) osif_io_in8( (addr) )
#define OSIF_IO_OUT16( addr, data ) osif_io_out16( (addr), (data) )
#define OSIF_IO_IN16( addr ) osif_io_in16( (addr) )
#define OSIF_IO_OUT32( addr, data ) osif_io_out32( (addr), (data) )
#define OSIF_IO_IN32( addr ) osif_io_in32( (addr) )

#define OSIF_MEM_OUT8( addr, data ) osif_mem_out8( (addr), (data) )
#define OSIF_MEM_IN8( addr ) osif_mem_in8( (addr) )
#define OSIF_MEM_OUT16( addr, data ) osif_mem_out16( (addr), (data) )
#define OSIF_MEM_IN16( addr ) osif_mem_in16( (addr) )
#define OSIF_MEM_OUT32( addr, data ) osif_mem_out32( (addr), (data) )
#define OSIF_MEM_IN32( addr ) osif_mem_in32( (addr) )


#if 0 /* test-only: irix */
#define OSIF_READB(base, reg)             (*(volatile UINT8* )((base)+(reg)))
#define OSIF_WRITEB(base, reg, data)      (*(volatile UINT8* )((base)+(reg))) = (data)
#endif

#define OSIF_READW(base, reg)             (*(volatile UINT16*)((base)+(reg)))
#define OSIF_WRITEW(base, reg, data)      (*(volatile UINT16*)((base)+(reg))) = (data)
#define OSIF_READLB(base, reg)            (*(volatile UINT32*)((base)+(reg)))
#define OSIF_WRITEL(base, reg, data)      (*(volatile UINT32*)((base)+(reg))) = (data)

#define OSIF_CPU2BE16(x) osif_cpu2be16(x)
#define OSIF_CPU2BE32(x) osif_cpu2be32(x)
#define OSIF_CPU2BE64(x) osif_cpu2be64(x)
#define OSIF_BE162CPU(x) osif_be162cpu(x)
#define OSIF_BE322CPU(x) osif_be322cpu(x)
#define OSIF_BE642CPU(x) osif_be642cpu(x)
#define OSIF_CPU2LE16(x) osif_cpu2le16(x)
#define OSIF_CPU2LE32(x) osif_cpu2le32(x)
#define OSIF_CPU2LE64(x) osif_cpu2le64(x)
#define OSIF_LE162CPU(x) osif_le162cpu(x)
#define OSIF_LE322CPU(x) osif_le322cpu(x)
#define OSIF_LE642CPU(x) osif_le642cpu(x)

#define OSIF_LOAD_FROM_FILE( d, f ) osif_load_from_file( d, f )

#define OSIF_DIV64_64(n, div) osif_div64_64(n, div)

#define OSIF_SER_OUT(str) osif_ser_out((str))

#define OSIF_MAKE_NTCAN_SERIAL(pszSerial, pulSerial) \
        osif_make_ntcan_serial(pszSerial, pulSerial)

/*
 *  Output-zones, use these defines together with OSIF_DPRINT to restrict output
 *  of messages to certain zones. Only zones marked with the lowest byte can produce
 *  output to the user in release-versions. All other versions are debug-only.
 */
#define OSIF_ZONE_NONE     0x00000000  /* no output                                    */
#define OSIF_ZONE_USR_INIT 0x00000001  /* driver load/unload  messages                 */
#define OSIF_ZONE_USR_INFO 0x00000002  /* detailed driver information and runtime msgs */
#define OSIF_ZONE_USR_RES3 0x00000004  /* reserved for future use                      */
#define OSIF_ZONE_USR_STAT 0x00000008  /* driver statistics on unload                  */
#define OSIF_ZONE_USR_RES5 0x00000010  /* reserved for future use                      */
#define OSIF_ZONE_USR_DBG  0x00000020  /* driver troubleshooting information           */
#define OSIF_ZONE_USR_RES7 0x00000040  /* reserved for future use                      */
#define OSIF_ZONE_USR_RES8 0x00000080  /* reserved for future use                      */

#define OSIF_ZONE_ESDCAN   0x00100000   /* messages from esdcan-layer                  */
#define OSIF_ZONE_NUC      0x00200000   /* messages from nucleus                       */
#define OSIF_ZONE_BOARD    0x00400000   /* messages from board-layer                   */
#define OSIF_ZONE_CTRL     0x00800000   /* messages from controller-layer              */
#define OSIF_ZONE_OSIF     0x01000000   /* messages from osif-layer                    */

#define OSIF_ZONE_FUNC     0x02000000   /* function entry/exit-messages                */
#define OSIF_ZONE_PARAM    0x04000000   /* parameter output                            */
#define OSIF_ZONE_IRQ      0x08000000   /* messages from ISR (USE THIS ZONE ALONE!!!)  */
#define OSIF_ZONE_BACKEND  0x10000000   /* output from backend run level               */
#define OSIF_ZONE_INIT     0x20000000   /* output from drivers initialization          */
#define OSIF_ZONE_WARN     0x40000000   /* warnings                                    */
#define OSIF_ZONE_ERROR    0x80000000   /* error-messages                              */

#ifndef OSIF_SUCCESS
#  define OSIF_SUCCESS                CANIO_SUCCESS
#endif
#ifndef OSIF_ERROR
#  define OSIF_ERROR                  CANIO_ERROR
#endif
#ifndef OSIF_INVALID_PARAMETER
#  define OSIF_INVALID_PARAMETER      CANIO_INVALID_PARAMETER
#endif
#ifndef OSIF_INVALID_HANDLE
#  define OSIF_INVALID_HANDLE         CANIO_INVALID_HANDLE
#endif
#ifndef OSIF_NET_NOT_FOUND
#  define OSIF_NET_NOT_FOUND          CANIO_NET_NOT_FOUND
#endif
#ifndef OSIF_INSUFFICIENT_RESOURCES
#  define OSIF_INSUFFICIENT_RESOURCES CANIO_INSUFFICIENT_RESOURCES
#endif
#ifndef OSIF_EAGAIN
#  define OSIF_EAGAIN                 CANIO_EAGAIN
#endif
#ifndef OSIF_EBUSY
#  define OSIF_EBUSY                  CANIO_EBUSY
#endif
#ifndef OSIF_RX_TIMEOUT
#  define OSIF_RX_TIMEOUT             CANIO_RX_TIMEOUT
#endif
#ifndef OSIF_TX_TIMEOUT
#  define OSIF_TX_TIMEOUT             CANIO_TX_TIMEOUT
#endif
#ifndef OSIF_TX_ERROR
#  define OSIF_TX_ERROR               CANIO_TX_ERROR
#endif
#ifndef OSIF_CONTR_OFF_BUS
#  define OSIF_CONTR_OFF_BUS          CANIO_CONTR_OFF_BUS
#endif
#ifndef OSIF_CONTR_BUSY
#  define OSIF_CONTR_BUSY             CANIO_CONTR_BUSY
#endif
#ifndef OSIF_CONTR_WARN
#  define OSIF_CONTR_WARN             CANIO_CONTR_WARN
#endif
#ifndef OSIF_OLDDATA
#  define OSIF_OLDDATA                CANIO_OLDDATA
#endif
#ifndef OSIF_NO_ID_ENABLED
#  define OSIF_NO_ID_ENABLED          CANIO_NO_ID_ENABLED
#endif
#ifndef OSIF_ID_ALREADY_ENABLED
#  define OSIF_ID_ALREADY_ENABLED     CANIO_ID_ALREADY_ENABLED
#endif
#ifndef OSIF_ID_NOT_ENABLED
#  define OSIF_ID_NOT_ENABLED         CANIO_ID_NOT_ENABLED
#endif
#ifndef OSIF_INVALID_FIRMWARE
#  define OSIF_INVALID_FIRMWARE       CANIO_INVALID_FIRMWARE
#endif
#ifndef OSIF_MESSAGE_LOST
#  define OSIF_MESSAGE_LOST           CANIO_MESSAGE_LOST
#endif
#ifndef OSIF_INVALID_HARDWARE
#  define OSIF_INVALID_HARDWARE       CANIO_INVALID_HARDWARE
#endif
#ifndef OSIF_PENDING_WRITE
#  define OSIF_PENDING_WRITE          CANIO_PENDING_WRITE
#endif
#ifndef OSIF_PENDING_READ
#  define OSIF_PENDING_READ           CANIO_PENDING_READ
#endif
#ifndef OSIF_INVALID_DRIVER
#  define OSIF_INVALID_DRIVER         CANIO_INVALID_DRIVER
#endif
#ifndef OSIF_WRONG_DEVICE_STATE
#  define OSIF_WRONG_DEVICE_STATE     CANIO_WRONG_DEVICE_STATE
#endif
#ifndef OSIF_OPERATION_ABORTED
#  define OSIF_OPERATION_ABORTED      CANIO_OPERATION_ABORTED
#endif
#ifndef OSIF_WRONG_DEVICE_STATE
#  define OSIF_WRONG_DEVICE_STATE     CANIO_WRONG_DEVICE_STATE
#endif
#ifndef OSIF_HANDLE_FORCED_CLOSE
#  define OSIF_HANDLE_FORCED_CLOSE    CANIO_HANDLE_FORCED_CLOSE
#endif
#ifndef OSIF_NOT_IMPLEMENTED
#  define OSIF_NOT_IMPLEMENTED        CANIO_NOT_IMPLEMENTED
#endif
#ifndef OSIF_NOT_SUPPORTED
#  define OSIF_NOT_SUPPORTED          CANIO_NOT_SUPPORTED
#endif
#ifndef OSIF_CONTR_ERR_PASSIVE
#  define OSIF_CONTR_ERR_PASSIVE      CANIO_CONTR_ERR_PASSIVE
#endif
#ifndef OSIF_EFAULT
#  define OSIF_EFAULT                 CANIO_EFAULT
#endif
#ifndef OSIF_EIO
#  define OSIF_EIO                    CANIO_EIO
#endif

/* macros used for printing 64 bit numbers */
#ifndef OSIF_PRIx64
#define OSIF_PRIx64 "llx"
#endif
#ifndef OSIF_PRId64
#define OSIF_PRId64 "lld"
#endif
#ifndef OSIF_PRIu64
#define OSIF_PRIu64 "llu"
#endif

#endif
