/* -*- linux-c -*-
 * FILE NAME osifi.h
 *           copyright 2002 by esd electronic system design GmbH
 *
 * BRIEF MODULE DESCRIPTION
 *           ...
 *
 *
 *
 * history:
 *
 *  $Log: osifi.h,v $
 *  Revision 1.56  2006/10/12 10:35:23  manuel
 *  Introduced inline assembler macros for swaps on x86
 *
 *  Revision 1.55  2006/02/14 09:17:46  matthias
 *  -don't use OSIF_CALLTYPE with regparm(0) on PPC (not suppoted by ppc compiler)
 *
 *  Revision 1.54  2005/12/08 11:33:41  andreas
 *  Added comment for kernel 2.2.x users
 *
 *  Revision 1.53  2005/11/02 07:00:40  andreas
 *  Added declaration of osif_make_ntcan_serial()
 *
 *  Revision 1.52  2005/10/06 07:43:03  matthias
 *  added 64bit swap macros
 *
 *  Revision 1.51  2005/09/23 15:28:37  manuel
 *  Removed OSIF_PCI_OFFSET=0xC0000000 for PPC
 *
 *  Revision 1.50  2005/07/27 15:50:51  andreas
 *  Corrected OSIF_IRQ_HANDLER_CALLTYPE definition
 *
 *  Revision 1.49  2005/07/13 08:57:04  matthias
 *  added OSIF_IRQ_HANDLER_CALLTYPE
 *
 *  Revision 1.48  2005/04/20 13:00:47  andreas
 *  Added osif_ser_out() (prints debug output to serial)
 *
 *  Revision 1.47  2004/10/29 15:28:53  matthias
 *  use UINT32 instead of u32
 *
 *  Revision 1.46  2004/10/29 14:54:11  matthias
 *  added osif_div64_64
 *
 *  Revision 1.45  2004/10/11 09:11:51  andreas
 *  Changes for linux kernel 2.6
 *   (swap functions, i/o functions, etc.)
 *  Added cvs log tag
 *  Sorted history
 *
 *  12.10.06 - Introduced inline assembler macros for swaps on x86  mk
 *  23.09.05 - Removed OSIF_PCI_OFFSET=0xC0000000 for PPC           mk
 *  13.05.04 - Added OSIF_CALLTYPE to timer functions               ab
 *  11.05.04 - Reworked interrupt handler stuff again
 *             Changed OSIF_EXTERN to extwern OSIF_CALLTYPE         ab
 *  07.05.04 - Changed definition of interrupt handler stuff to
 *               comply with kernel 2.6                             ab
 *  06.05.04 - changed include of modversions for kernel 2.6.x
 *             added asmlinkage to OSIF_EXTERN for kernel 2.6.x
 *             added OSIF_BIG_ENDIAN, etc. for kernel 2.6.x         ab
 *  01.12.03 - removed #ifdef in spinlock-structure                 ab
 *  26.11.03 - Removed OSIF_PCI_MALLOC warnings for non 405-cards   ab
 *  13.11.03 - Corrected debug-info in mutexes                      ab
 *  13.11.03 - OSIF_MUTEX is a sema now Old one is now
 *             OSIF_BUSY_MUTEX                                      mt
 *  05.11.03 - Removed IRQ-flags from mutex structure again         ab
 *  03.11.03 - added field for IRQ-flags in mutex-structure         ab
 *  27.10.03 - New output function osif_dprint (with verbose mask)  ab
 *  18.09.03 - Only one thread for all DPC's                        mt
 *  24.05.02 - first version                                        mf
 *
 */
/************************************************************************
 *
 *  Copyright (c) 1996 - 2002 by electronic system design gmbh
 *
 *  This software is copyrighted by and is the sole property of
 *  esd gmbh.  All rights, title, ownership, or other interests
 *  in the software remain the property of esd gmbh. This
 *  software may only be used in accordance with the corresponding
 *  license agreement.  Any unauthorized use, duplication, transmission,
 *  distribution, or disclosure of this software is expressly forbidden.
 *
 *  This Copyright notice may not be removed or modified without prior
 *  written consent of esd gmbh.
 *
 *  esd gmbh, reserves the right to modify this software without notice.
 *
 *  electronic system design gmbh          Tel. +49-511-37298-0
 *  Vahrenwalder Str 207                   Fax. +49-511-37298-68
 *  30165 Hannover                         http://www.esd-electronics.com
 *  Germany                                sales@esd-electronics.com
 *
 *************************************************************************/
/*! \file osifi.h
    \brief OSIF operating system wrapping header

    This file contains the OSIF prototypes and user macros.
    Any driver should call the OSIF_<xxx> macros instead of the
    osif_<xxx> functions.
*/

#ifndef __OSIFI_H__
#define __OSIFI_H__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/errno.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
# ifdef OSIF_CALLTYPE
#  undef OSIF_CALLTYPE
# endif
# if !defined(CONFIG_PPC)
#  define OSIF_CALLTYPE __attribute__((regparm(0))) /* asmlinkage */
# else
#  define OSIF_CALLTYPE
# endif
# if defined(CONFIG_X86)
#  define OSIF_ARCH_X86
#  define OSIF_LITTLE_ENDIAN
# endif
# if defined(CONFIG_PPC)
#  define OSIF_ARCH_PPC
#  define OSIF_BIG_ENDIAN
# endif
#endif

#include "sysdep.h"

#ifndef NULL
#define NULL ((void*)0)
#endif

#ifdef __KERNEL__

typedef struct __OSIF_SEMA  *OSIF_SEMA, *OSIF_MUTEX;

typedef struct __OSIF_SPINLOCK  *OSIF_BUSY_MUTEX, *OSIF_IRQ_MUTEX;

typedef struct __OSIF_DPC _OSIF_DPC, *OSIF_DPC;
struct __OSIF_DPC {
	OSIF_DPC            next;
	INT8                linked;
	INT8                running;
	INT8                destroy;
	INT8                reserved;
	OSIF_CALLTYPE VOID  (*func)(VOID *);
	VOID                *arg;
};

typedef struct __OSIF_TIMER  _OSIF_TIMER, *OSIF_TIMER;

#define OSIF_IRQ_HANDLER_CALLTYPE      OSIF_CALLTYPE
#define OSIF_IRQ_HANDLER_PARAMETER     void*
#define OSIF_IRQ_HANDLER(func,arg)     int (OSIF_IRQ_HANDLER_CALLTYPE func) ( void *arg )
#define OSIF_IRQ_HANDLER_P             int (OSIF_IRQ_HANDLER_CALLTYPE *) ( void *arg )
#define OSIF_IRQ_HANDLED               (1)  /* Equal to Linux2.6 IRQ_HANDLED */
#define OSIF_IRQ_HANDLED_NOT           (0)  /* Equal to Linux2.6 IRQ_NONE    */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
# define OSIF_IRQ_HANDLER_RETURN( arg ) return ((arg) != 0) /* Linux2.6: IRQ_RETVAL(x) */
#else
# define OSIF_IRQ_HANDLER_RETURN( arg ) return arg
#endif

typedef struct pci_dev *OSIF_PCIDEV;
#define OSIF_PCI_VADDR                 void *
#define OSIF_PCI_MADDR                 UINT32

/*
 *  Dear kernel 2.2.x user,
 *  we are not able to determine the exact version, when
 *  dma_addr_t was introduced into 2.2.x kernels.
 *  If the following typedef leads to an compiletime error,
 *  please comment the typedef.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
typedef u32 dma_addr_t;
#endif

typedef struct __OSIF_PCI_PADDR  _OSIF_PCI_PADDR, *OSIF_PCI_PADDR;

extern OSIF_CALLTYPE void   osif_io_out8(UINT8 *addr, UINT8 data);
extern OSIF_CALLTYPE UINT8  osif_io_in8(UINT8 *addr);
extern OSIF_CALLTYPE void   osif_io_out16(UINT16 *addr, UINT16 data);
extern OSIF_CALLTYPE UINT16 osif_io_in16(UINT16 *addr);
extern OSIF_CALLTYPE void   osif_io_out32(UINT32 *addr, UINT32 data);
extern OSIF_CALLTYPE UINT32 osif_io_in32(UINT32 *addr);

#define osif_mem_out8( addr, data )  (*(volatile UINT8* )(addr)) = (data)
#define osif_mem_in8( addr )         (*(volatile UINT8* )(addr))
#define osif_mem_out16( addr, data ) (*(volatile UINT16*)(addr)) = (data)
#define osif_mem_in16( addr )        (*(volatile UINT16*)(addr))
#define osif_mem_out32( addr, data ) (*(volatile UINT32*)(addr)) = (data)
#define osif_mem_in32( addr )        (*(volatile UINT32*)(addr))

#if defined(__GNUC__) && defined(CONFIG_X86)
/* too bad we cannot use 486 assembler here, because nucleus is compiled without knowledge
   of the system it is later used on */
#define osif_swap16(x) \
({ \
        register UINT16 __x; \
        asm ("xchg %b1,%h1" : "=q" (__x) : "0" ((UINT16)x)); \
        __x; \
})
#define osif_swap32(x) \
({ \
	register UINT32 __x; \
        asm ("xchg %b1,%h1; roll $16,%1; xchg %b1,%h1" : "=q" (__x) : "0" ((UINT32)x)); \
        __x; \
})
#define osif_swap64(x) \
({ \
	register UINT64 __x; \
        asm ("xchg %%ah,%%al; roll $16,%%eax; xchg %%ah,%%al; " \
             "xchg %%eax,%%edx; " \
             "xchg %%ah,%%al; roll $16,%%eax; xchg %%ah,%%al" : "=A" (__x) : "0" ((UINT64)x)); \
        __x; \
})

#else    /* sorry, no inline assembler macros for you */

#define osif_swap16(x) \
({ \
	UINT16 __x = (x); \
	((UINT16)( \
		(((UINT16)(__x) & (UINT16)0x00ffU) << 8) | \
		(((UINT16)(__x) & (UINT16)0xff00U) >> 8) )); \
})
#define osif_swap32(x) \
({ \
	UINT32 __x = (x); \
	((UINT32)( \
		(((UINT32)(__x) & (UINT32)0x000000ffUL) << 24) | \
		(((UINT32)(__x) & (UINT32)0x0000ff00UL) <<  8) | \
		(((UINT32)(__x) & (UINT32)0x00ff0000UL) >>  8) | \
		(((UINT32)(__x) & (UINT32)0xff000000UL) >> 24) )); \
})
#define osif_swap64(x) \
({ \
	UINT64 __x = (x); \
	((UINT64)( \
		(((UINT64)(__x) & (UINT64)0x00000000000000ffULL) << (8*7)) | \
		(((UINT64)(__x) & (UINT64)0x000000000000ff00ULL) << (8*5)) | \
		(((UINT64)(__x) & (UINT64)0x0000000000ff0000ULL) << (8*3)) | \
		(((UINT64)(__x) & (UINT64)0x00000000ff000000ULL) << (8*1)) | \
		(((UINT64)(__x) & (UINT64)0x000000ff00000000ULL) >> (8*1)) | \
		(((UINT64)(__x) & (UINT64)0x0000ff0000000000ULL) >> (8*3)) | \
		(((UINT64)(__x) & (UINT64)0x00ff000000000000ULL) >> (8*5)) | \
		(((UINT64)(__x) & (UINT64)0xff00000000000000ULL) >> (8*7)) )); \
})
#endif

#ifdef OSIF_LITTLE_ENDIAN
#define osif_cpu2be16(x) osif_swap16(x)
#define osif_cpu2be32(x) osif_swap32(x)
#define osif_cpu2be64(x) osif_swap64(x)
#define osif_be162cpu(x) osif_swap16(x)
#define osif_be322cpu(x) osif_swap32(x)
#define osif_be642cpu(x) osif_swap64(x)
#define osif_cpu2le16(x) (x)
#define osif_cpu2le32(x) (x)
#define osif_cpu2le64(x) (x)
#define osif_le162cpu(x) (x)
#define osif_le322cpu(x) (x)
#define osif_le642cpu(x) (x)
#else
#define osif_cpu2be16(x) (x)
#define osif_cpu2be32(x) (x)
#define osif_cpu2be64(x) (x)
#define osif_be162cpu(x) (x)
#define osif_be322cpu(x) (x)
#define osif_be642cpu(x) (x)
#define osif_cpu2le16(x) osif_swap16(x)
#define osif_cpu2le32(x) osif_swap32(x)
#define osif_cpu2le64(x) osif_swap64(x)
#define osif_le162cpu(x) osif_swap16(x)
#define osif_le322cpu(x) osif_swap32(x)
#define osif_le642cpu(x) osif_swap64(x)
#endif

extern OSIF_CALLTYPE INT32 osif_attach( VOID );
extern OSIF_CALLTYPE INT32 osif_detach( VOID );

extern OSIF_CALLTYPE INT32 osif_malloc(UINT32 size, VOID **p);
extern OSIF_CALLTYPE INT32 osif_free(VOID *ptr);
extern OSIF_CALLTYPE INT32 osif_memset(VOID *ptr, UINT32 val, UINT32 size);
extern OSIF_CALLTYPE INT32 osif_memcpy(VOID *dst, VOID *src, UINT32 size);
extern OSIF_CALLTYPE INT32 osif_memcmp(VOID *dst, VOID *src, UINT32 size);

extern OSIF_CALLTYPE VOID set_output_mask(UINT32 mask);

#ifdef OSIF_OS_RTAI
# ifdef BOARD_pci405fw
extern OSIF_CALLTYPE INT32 pci405_printk(const char *fmt, ...);
extern OSIF_CALLTYPE INT32 pci405_printk_init( VOID );
extern OSIF_CALLTYPE INT32 pci405_printk_cleanup( VOID );
#  define osif_print(fmt... ) pci405_printk( ## fmt )
# else
#  define osif_print(fmt... ) rt_printk( ## fmt )
# endif
#else
extern OSIF_CALLTYPE INT32 osif_print(const INT8 *fmt, ...);
#endif
extern OSIF_CALLTYPE INT32 osif_dprint(UINT32 output_mask, const INT8 *fmt, ...);

extern OSIF_CALLTYPE INT32 osif_snprintf(UINT8 *str, INT32 size, const INT8 *format, ...);

extern OSIF_CALLTYPE INT32 osif_sema_create(OSIF_SEMA *s, INT32 init);
extern OSIF_CALLTYPE INT32 osif_sema_destroy(OSIF_SEMA *s);
extern OSIF_CALLTYPE INT32 osif_sema_down(OSIF_SEMA *s);
extern OSIF_CALLTYPE INT32 osif_sema_down_interruptible(OSIF_SEMA *s);
extern OSIF_CALLTYPE INT32 osif_sema_up(OSIF_SEMA *s);

extern OSIF_CALLTYPE INT32 osif_mutex_create(OSIF_MUTEX *m);
extern OSIF_CALLTYPE INT32 osif_mutex_destroy(OSIF_MUTEX *m);
extern OSIF_CALLTYPE INT32 osif_mutex_lock(OSIF_MUTEX *m);
extern OSIF_CALLTYPE INT32 osif_mutex_unlock(OSIF_MUTEX *m);

extern OSIF_CALLTYPE INT32 osif_irq_mutex_create( OSIF_IRQ_MUTEX  *im);
extern OSIF_CALLTYPE INT32 osif_irq_mutex_destroy( OSIF_IRQ_MUTEX  *im);
extern OSIF_CALLTYPE INT32 osif_irq_mutex_lock( OSIF_IRQ_MUTEX  *im);
extern OSIF_CALLTYPE INT32 osif_irq_mutex_unlock( OSIF_IRQ_MUTEX  *im);
extern OSIF_CALLTYPE INT32 osif_in_irq_mutex_lock( OSIF_IRQ_MUTEX  *im);
extern OSIF_CALLTYPE INT32 osif_in_irq_mutex_unlock( OSIF_IRQ_MUTEX  *im);

extern OSIF_CALLTYPE INT32 osif_sleep(INT32 ms);

extern OSIF_CALLTYPE INT32 osif_dpc_create( OSIF_DPC  *dpc, OSIF_CALLTYPE VOID(*func)(VOID *), VOID *arg );
extern OSIF_CALLTYPE INT32 osif_dpc_destroy( OSIF_DPC *dpc );
extern OSIF_CALLTYPE INT32 osif_dpc_trigger( OSIF_DPC *dpc );

extern OSIF_CALLTYPE INT32 osif_timer_create( OSIF_TIMER  *timer, OSIF_CALLTYPE VOID(*func)(VOID *), VOID *arg );
extern OSIF_CALLTYPE INT32 osif_timer_destroy( OSIF_TIMER *timer );
extern OSIF_CALLTYPE INT32 osif_timer_set( OSIF_TIMER *timer, UINT32 ms );
extern OSIF_CALLTYPE INT32 osif_timer_get( OSIF_TIMER *t, UINT32 *ms );

#if 0
extern OSIF_CALLTYPE INT32 osif_irq_attach( UINT32 irq, OSIF_IRQ_HANDLER_ARG, UINT8 *name, VOID *arg );
extern OSIF_CALLTYPE INT32 osif_irq_detach( UINT32 irq, VOID *arg );
#endif

extern OSIF_CALLTYPE UINT64 osif_ticks( VOID );
extern OSIF_CALLTYPE UINT64 osif_tick_frequency( VOID );

extern OSIF_CALLTYPE INT32 osif_pci_malloc(OSIF_PCIDEV pcidev, OSIF_PCI_VADDR *vaddr,
                                           OSIF_PCI_PADDR *paddr, OSIF_PCI_MADDR *maddr, UINT32 size);
extern OSIF_CALLTYPE INT32 osif_pci_free(OSIF_PCIDEV pcidev, OSIF_PCI_VADDR vaddr,
                                         OSIF_PCI_PADDR *paddr, OSIF_PCI_MADDR maddr, UINT32 size);
extern OSIF_CALLTYPE UINT32 osif_pci_get_phy_addr(OSIF_PCI_PADDR paddr);

extern OSIF_CALLTYPE INT32 osif_pci_read_config_byte(OSIF_PCIDEV pcidev, INT32 offs, UINT8 *ptr);
extern OSIF_CALLTYPE INT32 osif_pci_read_config_word(OSIF_PCIDEV pcidev, INT32 offs, UINT16 *ptr);
extern OSIF_CALLTYPE INT32 osif_pci_read_config_long(OSIF_PCIDEV pcidev, INT32 offs, UINT32 *ptr);
extern OSIF_CALLTYPE INT32 osif_pci_write_config_byte(OSIF_PCIDEV pcidev, INT32 offs, UINT8 val);
extern OSIF_CALLTYPE INT32 osif_pci_write_config_word(OSIF_PCIDEV pcidev, INT32 offs, UINT16 val);
extern OSIF_CALLTYPE INT32 osif_pci_write_config_long(OSIF_PCIDEV pcidev, INT32 offs, UINT32 val);

extern OSIF_CALLTYPE INT32 osif_load_from_file( UINT8 *destaddr, UINT8 *filename );

extern OSIF_CALLTYPE VOID osif_ser_out(UINT8 *str);

extern OSIF_CALLTYPE UINT32 osif_div64_32(UINT64 *n, UINT32 base);

#define osif_div64_64(n, div)	({	        \
	UINT64 __n = (n);			\
	UINT32 __d = (div);			\
	UINT32 __q, __r;			\
	if ((__n >> 32) == 0) {	                \
		__q = (UINT32)__n / __d;	\
		__r = (UINT32)__n - __q * __d;	\
		(n) = __q;			\
        } else if (((div) >> 32) == 0x00000001) { \
        /* very inacurate for numbers 0x100000000<div<0x200000000, maybe better: */\
        /*} else if ((div) == 0x100000000ULL) {*/\
		(n) = __n >> 32;                \
                __r = (UINT32)__n;              \
	} else {				\
		__r = osif_div64_32(&__n, __d);	\
		(n) = __n;			\
	}					\
	__r;					\
})

/*
 * If we get a serial it is assumed that it has the format CCNNNNNN where
 * C is a character from 'A'to 'P' and N is a number from '0' to '9'
 * This is converted into an 32-bit integer with bit 31-28 as alphabetical
 * offset to 'A' of the first char in serial, bit 27-24 as alphabetical
 * offset to 'A' of the second char in serial and the remaining 24 bits
 * as serial. So we have an effective range from AA000000 to PPFFFFFF.
 */
extern OSIF_CALLTYPE INT32 osif_make_ntcan_serial(const INT8 *pszSerial, UINT32 *pulSerial);


#ifndef OSIF_PCI_OFFSET
# define OSIF_PCI_OFFSET 0
#endif

/*
 * If OSIF_TICK_FREQ is not defined here, UINT64 osif_tick_frequency()
 * must be implemented in osif.c
 */
#ifdef OSIF_ARCH_X86
#define OSIF_TICK_FREQ              1000000LL
#endif

#define OSIF_READB(base, reg)             (*(volatile UINT8* )((base)+(reg)))
#define OSIF_WRITEB(base, reg, data)      (*(volatile UINT8* )((base)+(reg))) = (data)


#define OSIF_KERNEL
#endif /* __KERNEL__ */

#define OSIF_ERRNO_BASE             0x00000100  /* base for esd-codes */

#endif

